#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
import os
import glob
from itertools import product
import yaml

from .model import Model, ModelFormats
from .group import ModelGroup

from mlonmcu.logging import get_logger

logger = get_logger()


def get_model_directories(context):
    dirs = context.environment.paths["models"]
    dirs = [d.path for d in dirs]
    return dirs


def find_metadata(directory, model_name=None):
    possible_basenames = ["model", "metadata", "definition"]
    possible_extensions = ["yaml", "yml"]
    directory = Path(directory)
    dirname = directory.name
    if dirname not in possible_basenames:
        possible_basenames.insert(0, dirname)
    if model_name and model_name not in possible_basenames:
        possible_basenames.insert(0, model_name)

    for combination in product(possible_basenames, possible_extensions):
        filename = f"{combination[0]}.{combination[1]}"
        fullpath = directory / filename
        if fullpath.is_file():
            return fullpath
            # logger.debug("Found match. Ignoring other files")
    return None


def list_models(directory, depth=1, formats=None, config=None):  # TODO: get config from environment or cmdline!
    config = config if config is not None else {}
    formats = formats if formats else [ModelFormats.TFLITE]
    assert len(formats) > 0, "No formats provided for model lookup"
    models = []
    for fmt in formats:
        if depth != 1:
            raise NotImplementedError  # TODO: implement for arm ml zoo
            # define all allowed extensions + search recusively (with limit?)
            # list(Path(".").rglob(f"*.{ext}")) for ext in allowed_ext
        if not os.path.isdir(directory):
            logger.debug("Not a directory: %s", str(directory))
            return []
        subdirs = [Path(directory) / o for o in os.listdir(directory) if os.path.isdir(os.path.join(directory, o))]
        for subdir in subdirs:
            dirname = subdir.name
            if dirname.startswith("."):
                # logger.debug("Skipping hidden directory: %s", str(dirname))
                continue
            exts = fmt.extensions
            for ext in exts:
                main_model = subdir / f"{dirname}.{ext}"
                if os.path.exists(main_model):
                    main_model = f"{dirname}/{dirname}"
                else:
                    main_model = None
                submodels = []
                for filename in glob.glob(str(subdir / f"*.{ext}")):
                    basename = "".join(Path(filename).name.split(".")[:-1])
                    submodels.append(f"{dirname}/{basename}")

                if len(submodels) == 1:
                    main_model = submodels[0]

                if main_model:
                    submodels.remove(main_model)

                    main_base = main_model.split("/")[-1]
                    main_config = {}
                    main_metadata_path = find_metadata(Path(directory) / dirname, model_name=main_base)
                    if main_metadata_path:
                        main_config[f"{main_base}.metadata_path"] = main_metadata_path
                    main_config.update(config)

                    models.append(
                        Model(
                            main_base,
                            [Path(directory) / f"{main_model}.{ext}"],
                            config=main_config,
                            alt=main_model,
                            formats=[fmt],
                        )
                    )

                for submodel in submodels:
                    sub_base = submodel.split("/")[-1]
                    submodel_config = {}
                    submodel_metadata_path = find_metadata(Path(directory) / dirname, model_name=sub_base)
                    if submodel_metadata_path:
                        submodel_config[f"{sub_base}.metadata_path"] = submodel_metadata_path
                    submodel_config.update(config)
                    models.append(
                        Model(
                            submodel,
                            [Path(directory) / f"{submodel}.{ext}"],
                            config=submodel_config,
                            formats=[fmt],
                        )
                    )

    return models


def list_modelgroups(directory):
    if not os.path.isdir(directory):
        logger.debug("Not a directory: %s", str(directory))
        return []
    groups = []
    directory = Path(directory)
    possible_basenames = ["modelgroups", "groups"]
    possible_extensions = ["yaml", "yml"]
    for combination in product(possible_basenames, possible_extensions):
        filename = f"{combination[0]}.{combination[1]}"
        fullpath = directory / filename
        if fullpath.is_file():
            logger.debug("Found match. Ignoring other files")
            with open(fullpath, "r") as yamlfile:
                try:
                    content = yaml.safe_load(yamlfile)
                    for groupname, groupmodels in content.items():
                        assert isinstance(groupmodels, list), "Modelgroups should be defined as a YAML list"
                        modelgroup = ModelGroup(groupname, groupmodels)
                        groups.append(modelgroup)
                except yaml.YAMLError as err:
                    raise RuntimeError("Could not open YAML file") from err
            break
    return groups


def lookup_models_and_groups(directories, formats, config=None):
    all_models = []
    all_groups = []
    duplicates = {}
    group_duplicates = {}
    for directory in directories:
        models = list_models(directory, formats=formats, config=config)
        if len(all_models) == 0:
            all_models = models
        else:
            all_model_names = [m.name for m in all_models]
            for model in models:
                name = model.name
                if name in all_model_names:
                    if name in duplicates:
                        duplicates[name] += 1
                    else:
                        duplicates[name] = 1
                else:
                    all_models.append(model)
        groups = list_modelgroups(directory)
        if len(all_groups) == 0:
            all_groups = groups
        else:
            all_group_names = [g.name for g in all_groups]
            for group in groups:
                name = group.name
                if name in all_group_names:
                    if name in group_duplicates:
                        group_duplicates[name] += 1
                    else:
                        group_duplicates[name] = 1
                else:
                    all_groups.append(group)

    all_models = sorted(all_models, key=lambda x: x.name)

    return all_models, all_groups, duplicates, group_duplicates


def print_paths(directories):
    print("Paths:")
    for directory in directories:
        exists = os.path.exists(directory)
        print("    " + str(directory), end="\n" if exists else " (skipped)\n")
        # if not exists:
        #     directories.remove(directory)
    print()


def print_models(models, duplicates=[], detailed=False):
    print("Models:")
    for model in models:
        name = model.name
        path = model.paths[0]
        meta = "available" if model.metadata is not None else "not available"
        print("    " + name, end="")
        if name in duplicates:
            num = duplicates[name]
            print(f" ({num} duplicates)")
        else:
            print()
        if detailed:
            print(f"        Path: {path}")
            print(f"        Metadata: {meta}")
    print()


def print_groups(groups, all_models=[], duplicates=[], detailed=False):
    print("Groups:")
    for group in groups:
        name = group.name
        models = group.models
        all_model_names = [m.name for m in all_models]
        for model in models:
            if model in all_model_names:
                # Groupname conflicts with modelname
                group.name = "~" + group.name
        size = len(models)
        print(f"    {name} [{size} models]", end="")
        if name in duplicates:
            num = duplicates[name]
            print(f" ({num} duplicates)")
        else:
            print()

        if detailed:
            groupmodels = " ".join(models)
            print(f"        Models: {groupmodels}")
            print()


def print_summary(context, detailed=False):
    # TODO: get from context!
    formats = [ModelFormats.TFLITE]

    directories = get_model_directories(context)

    models, groups, duplicates, group_duplicates = lookup_models_and_groups(directories, formats)

    print("Models Summary\n")
    print_paths(directories)
    print_models(models, duplicates=duplicates, detailed=detailed)
    print_groups(groups, duplicates=group_duplicates, all_models=models, detailed=detailed)


def lookup_models(names, frontends=None, config=None, context=None):
    if frontends is None:
        assert context is not None
        # TODO: Get defaults frontends from environment (with no config/features)
        raise NotImplementedError
        # frontends = ?
    # allowed_ext = [frontend.fmt.extension for frontend in frontends]
    allowed_fmts = list(dict.fromkeys(sum([frontend.input_formats for frontend in frontends], [])))  # Remove duplicates
    allowed_exts = [fmt.extension for fmt in allowed_fmts]  # There should nit be duplicates

    if context:
        directories = get_model_directories(context)
        models, _, _, _ = lookup_models_and_groups(directories, allowed_fmts, config=config)
    else:
        models = []
    model_names = [model.name for model in models]

    hints = []
    for name in names:
        filepath = Path(name)
        real_name = filepath.stem
        ext = filepath.suffix[1:]
        if len(ext) > 0 and filepath.is_file():  # Explicit file
            assert (
                ext in allowed_exts
            ), f"Unsupported file extension for model which was explicitly passed by path: {ext}"
            paths = [filepath]
            metadata_path = find_metadata(filepath.parent.resolve(), model_name=real_name)
            if metadata_path:
                config[f"{real_name}.metadata_path"] = metadata_path
            hint = Model(real_name, paths, config=config, formats=[ModelFormats.from_extension(ext)])
            # TODO: look for metadata
            hints.append(hint)
        else:
            assert context is not None, "Context is required for passing models by name"
            assert len(model_names) > 0, "List of available models is empty"
            assert name in model_names, f"Could not find a model matching the name: {name}"
            index = model_names.index(name)
            hint = models[index]
            hints.append(hint)

    return hints


def apply_modelgroups(models, context=None):
    assert context is not None
    directories = get_model_directories(context)

    groups = {}
    for directory in directories:
        for group in list_modelgroups(directory):
            if group.name in groups and groups[group.name] != group.models:
                raise RuntimeError(
                    f"The model group '{group.name}' has conflicting definitions. Used the following directories:"
                    f" {directories}"
                )
            groups[group.name] = group.models

    out_models = []
    for model in models:
        out_models.extend(groups.get(model, [model]))
    return list(dict.fromkeys(out_models))  # Drop duplicates


def map_frontend_to_model(model, frontends, backend=None):
    model_fmts = model.formats
    backend_fmts = backend.get_supported_formats() if backend else []
    for frontend in frontends:
        ins = frontend.input_formats
        outs = frontend.output_formats
        assert len(ins) > 0
        # TODO: instead of picking the first frontend which at least one match, determine the highest overlap
        if any(in_fmt in model_fmts for in_fmt in ins):
            if len(backend_fmts) == 0 or any(out_fmt in backend_fmts for out_fmt in outs):
                new_model = model  # TODO: deepcopy?
                for i, fmt in enumerate(model_fmts):
                    if fmt not in ins:  # Only keep those formqts which are supported by the chosen frontend
                        path = model.paths[model_fmts.index(fmt)]
                        new_model.formats.remove(fmt)
                        new_model.paths.remove(path)
                return new_model, frontend
    raise RuntimeError(f"Unable to find a suitable frontend for model '{model.name}'")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from mlonmcu.target.arm.util import resolve_cpu_features


@pytest.mark.parametrize("model", ["cortex-m0", "cortex-m0plus", "cortex-m1", "cortex-m3", "cortex-m23"])
def test_target_arm_resolve_cpu_features1(model):
    # m0, m0plus, m1, m3, m23
    DSP = False
    MVE = False
    FP = False
    FP_DP = False
    assert resolve_cpu_features(model, None, None, None, None) == (model, "soft", "auto")
    assert resolve_cpu_features(model, FP, FP_DP, DSP, MVE) == (model, "soft", "auto")
    # assert resolve_cpu_features("cortex-m0", enable_fp, enable_fp_dp, enable_dsp, enable_mve) == (cpu, abi, fpu)


@pytest.mark.parametrize("model", ["cortex-m4"])
def test_target_arm_resolve_cpu_features2(model):
    DSP = True
    MVE = False
    FP = True
    FP_DP = False
    assert resolve_cpu_features(model, None, None, None, None) == (model, "hard", "auto")
    assert resolve_cpu_features(model + "+nofp", None, None, None, None) == (model + "+nofp", "soft", "auto")
    assert resolve_cpu_features(model, FP, FP_DP, DSP, MVE) == (model, "hard", "auto")
    assert resolve_cpu_features(model, False, FP_DP, DSP, MVE) == (model + "+nofp", "soft", "auto")


@pytest.mark.parametrize("model", ["cortex-m33", "cortex-m35p"])
def test_target_arm_resolve_cpu_features3(model):
    DSP = True
    MVE = False
    FP = True
    FP_DP = False
    assert resolve_cpu_features(model, None, None, None, None) == (model, "hard", "auto")
    assert resolve_cpu_features(model + "+nofp", None, None, None, None) == (model + "+nofp", "soft", "auto")
    assert resolve_cpu_features(model + "+nodsp", None, None, None, None) == (model + "+nodsp", "hard", "auto")
    assert resolve_cpu_features(model + "+nofp+nodsp", None, None, None, None) == (
        model + "+nofp+nodsp",
        "soft",
        "auto",
    )
    assert resolve_cpu_features(model, FP, FP_DP, DSP, MVE) == (model, "hard", "auto")
    assert resolve_cpu_features(model, False, FP_DP, DSP, MVE) == (model + "+nofp", "soft", "auto")
    assert resolve_cpu_features(model, FP, FP_DP, False, MVE) == (model + "+nodsp", "hard", "auto")
    assert resolve_cpu_features(model, False, FP_DP, False, MVE) == (model + "+nofp+nodsp", "soft", "auto")


@pytest.mark.parametrize("model", ["cortex-m7"])
def test_target_arm_resolve_cpu_features4(model):
    DSP = True
    MVE = False
    FP = True
    FP_DP = True
    assert resolve_cpu_features(model, None, None, None, None) == (model, "hard", "auto")
    assert resolve_cpu_features(model + "+nofp", None, None, None, None) == (model + "+nofp", "soft", "auto")
    assert resolve_cpu_features(model + "+nofp.dp", None, None, None, None) == (model + "+nofp.dp", "hard", "auto")
    assert resolve_cpu_features(model + "+nofp+nofp.dp", None, None, None, None) == (
        model + "+nofp+nofp.dp",
        "soft",
        "auto",
    )
    assert resolve_cpu_features(model, FP, FP_DP, DSP, MVE) == (model, "hard", "auto")
    assert resolve_cpu_features(model, False, False, DSP, MVE) == (model + "+nofp", "soft", "auto")
    assert resolve_cpu_features(model, FP, False, DSP, MVE) == (model + "+nofp.dp", "hard", "auto")


@pytest.mark.parametrize("model", ["cortex-m55"])
def test_target_arm_resolve_cpu_features5(model):
    DSP = True
    MVE = True
    FP = True
    FP_DP = True
    assert resolve_cpu_features(model, None, None, None, None) == (model, "hard", "auto")
    assert resolve_cpu_features(model + "+nofp", None, None, None, None) == (model + "+nofp", "soft", "auto")
    assert resolve_cpu_features(model + "+nodsp+nomve", None, None, None, None) == (
        model + "+nodsp+nomve",
        "hard",
        "auto",
    )
    assert resolve_cpu_features(model + "+nofp+nodsp+nomve", None, None, None, None) == (
        model + "+nofp+nodsp+nomve",
        "soft",
        "auto",
    )
    assert resolve_cpu_features(model + "+nomve", None, None, None, None) == (model + "+nomve", "hard", "auto")
    assert resolve_cpu_features(model, FP, FP_DP, DSP, MVE) == (model, "hard", "auto")
    assert resolve_cpu_features(model, False, False, DSP, MVE) == (model + "+nofp", "soft", "auto")
    assert resolve_cpu_features(model, FP, FP_DP, False, False) == (model + "+nodsp+nomve", "hard", "auto")
    assert resolve_cpu_features(model, False, False, False, False) == (model + "+nofp+nodsp+nomve", "soft", "auto")
    assert resolve_cpu_features(model, FP, FP_DP, DSP, False) == (model + "+nomve", "hard", "auto")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for the tune stage."""

from mlonmcu.cli.common import kickoff_runs
from mlonmcu.cli.build import add_build_options, handle as handle_build
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.session.run import RunStage


def get_parser(subparsers, parent=None):
    """ "Define and return a subparser for the tune subcommand."""
    parser = subparsers.add_parser(
        "tune",
        description="Tune model using the ML on MCU flow.",
        parents=[parent] if parent else [],
        add_help=(parent is None),
    )
    parser.set_defaults(flow_func=handle)
    add_build_options(parser)
    return parser


def handle(args, ctx=None):
    if ctx:
        handle_build(args, ctx, require_target=True)
    else:
        # args.features.append("autotune")  # TODO: enable autotuning automatically?
        # args.features.append("autotuned")  # ?
        with MlonMcuContext(path=args.home, deps_lock="read") as context:
            handle_build(args, context, require_target=True)
            kickoff_runs(args, RunStage.TUNE, context)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU ARM Cortex-M utilities"""

DSP_TARGETS = ["cortex-m4", "cortex-m7", "cortex-m33", "cortex-m35p", "cortex-m55"]

MVE_TARGETS = ["cortex-m55"]

FP_TARGETS = ["cortex-m4", "cortex-m7", "cortex-m33", "cortex-m35p", "cortex-m55"]
FP_DP_TARGETS = ["cortex-m7", "cortex-m55"]

ALLOWED_ATTRS = {
    "cortex-m0": [],
    "cortex-m0plus": [],
    "cortex-m1": [],
    "cortex-m3": [],
    "cortex-m4": ["+nofp"],
    "cortex-m7": ["+nofp", "+nofp.dp"],
    "cortex-m23": [],
    "cortex-m33": ["+nodsp", "+nofp"],
    "cortex-m35p": ["+nodsp", "+nofp"],
    "cortex-m55": ["+nomve", "+nomve.fp", "+nodsp", "+nofp"],
}


def resolve_cpu_features(model, enable_fp=None, enable_fp_dp=None, enable_dsp=None, enable_mve=None):
    cpu = model
    fpu = "auto"

    splitted = cpu.split("+")
    base = splitted[0]
    attrs = [f"+{attr}" for attr in splitted[1:]]

    if enable_fp is None:
        enable_fp = base in FP_TARGETS and ("+nofp" not in ALLOWED_ATTRS[base] or "+nofp" not in attrs)
    if enable_fp_dp is None:
        if enable_fp:
            enable_fp_dp = base in FP_DP_TARGETS and ("+nofp.dp" not in ALLOWED_ATTRS[base] or "+nofp.dp" not in attrs)
        else:
            enable_fp_dp = False
    if enable_dsp is None:
        enable_dsp = base in DSP_TARGETS and ("+nodsp" not in ALLOWED_ATTRS[base] or "+nodsp" not in attrs)
    if enable_mve is None:
        enable_mve = base in MVE_TARGETS and ("+nomve" not in ALLOWED_ATTRS[base] or "+nomve" not in attrs)

    float_abi = "hard" if enable_fp else "soft"

    if enable_fp_dp:
        assert enable_fp, "ARM double precision floating point supports requires fp extension"
    if enable_mve:
        assert enable_dsp, "ARM MVEI extension implies DSP extension"

    if len(attrs) > 0:
        # User defined features explicitly, check for validity
        if enable_fp and ("+nofp" not in ALLOWED_ATTRS[base] or "+nofp" not in attrs):
            assert base in FP_TARGETS, f"Chosen cpu '{cpu}' does not support floating point"
        if enable_fp_dp and ("+nofp.dp" not in ALLOWED_ATTRS[base] or "+nofp.dp" not in attrs):
            assert base in FP_DP_TARGETS, f"Chosen cpu '{cpu}' does not support double precison floating point"
        if enable_dsp and ("+nodsp" not in ALLOWED_ATTRS[base] or "+nodsp" not in attrs):
            assert base in DSP_TARGETS, f"Chosen cpu '{cpu}' does not support dsp instructions"
        if enable_mve and ("+nomve" not in ALLOWED_ATTRS[base] or "+nomve" not in attrs):
            assert base in MVE_TARGETS, f"Chosen cpu '{cpu}' does not support mve (helium) instructions"
    else:
        # Define features automatically
        if enable_fp:
            assert base in FP_TARGETS, f"Chosen cpu '{cpu}' does not support floating point"
        else:
            if base in FP_TARGETS:
                assert "+nofp" in ALLOWED_ATTRS[base], f"Floating point support can not be turned off for cpu '{cpu}'"
                attrs.append("+nofp")
        if enable_fp_dp:
            assert base in FP_DP_TARGETS, f"Chosen cpu '{cpu}' does not support double precision floating point"
        else:
            if base in FP_DP_TARGETS and enable_fp:
                assert (
                    "+nofp.dp" in ALLOWED_ATTRS[base]
                ), f"Double precision floating point support can not be turned off for cpu '{cpu}'"
                attrs.append("+nofp.dp")
        if enable_dsp:
            assert base in DSP_TARGETS, f"Chosen cpu '{cpu}' does not support dsp instructions"
        else:
            if base in DSP_TARGETS:
                assert "+nodsp" in ALLOWED_ATTRS[base], f"DSP extension can not be turned off for cpu '{cpu}'"
                attrs.append("+nodsp")
        if enable_mve:
            assert base in MVE_TARGETS, f"Chosen cpu '{cpu}' does not support mve (helium) instructions"
        else:
            if base in MVE_TARGETS:
                assert "+nomve" in ALLOWED_ATTRS[base], f"MVE (helium) extension can not be turned of for cpu '{cpu}'"
                attrs.append("+nomve")

    cpu = "".join([base] + attrs)
    return cpu, float_abi, fpu



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
from typing import Tuple

from .backend import TFLMBackend
from mlonmcu.config import str2bool, str2list
from mlonmcu.flow.backend import main
from mlonmcu.artifact import Artifact, ArtifactFormat


# TODO: move to another place
def make_hex_array(data):
    out = ""
    for x in data:
        out += "0x{:02x}, ".format(x)
    return out


class TFLMICodegen:
    def __init__(self):
        pass

    def makeCustomOpPrototypes(self, custom_ops):
        out = "namespace tflite {\n"
        for op in custom_ops:
            op_name = op
            op_reg = op
            if "|" in op:
                op_name, op_reg = op.split("|")[:2]
            out += "extern TfLiteRegistration *" + op_reg + "(void);\n"
        out += "}  //namespace tflite\n"
        return out

    def make_op_registrations(self, ops, custom_ops, reporter=True):
        out = (
            "static tflite::MicroMutableOpResolver<" + str(len(ops) + len(custom_ops)) + "> resolver(error_reporter);\n"
            if reporter
            else "static tflite::MicroMutableOpResolver<" + str(len(ops) + len(custom_ops)) + "> resolver;\n"
        )
        for op in ops:
            if reporter:
                out += (
                    "  if (resolver.Add"
                    + op
                    + '() != kTfLiteOk) {\n    error_reporter->Report("Add'  # TODO: replace with new logger
                    + op
                    + '() failed");\n    return 1;\n  }\n'
                )
            else:
                # TODO
                out += "  if (resolver.Add" + op + "() != kTfLiteOk) {\n    return 1;\n  }\n"
        for op in custom_ops:
            op_name = op
            op_reg = op
            if "|" in op:
                op_name, op_reg = op.split("|")[:2]
            if reporter:
                out += (
                    '  if (resolver.AddCustom("'
                    + op_name
                    + '", tflite::'
                    + op_reg
                    + '()) != kTfLiteOk) {\n    error_reporter->Report("AddCustom'  # TODO: replace with new logger
                    + op_name
                    + '() failed");\n    return 1;\n  }\n'
                )
            else:
                # TODO
                out += (
                    '  if (resolver.AddCustom("'
                    + op_name
                    + '", tflite::'
                    + op_reg
                    + "()) != kTfLiteOk) {\n    return 1;\n  }\n"
                )
        return out

    def generate_header(self, prefix="model"):
        upper_prefix = prefix.upper()
        code = f"""
// This file is generated. Do not edit.
#ifndef {upper_prefix}_GEN_H
#define {upper_prefix}_GEN_H

#include <stddef.h>

int {prefix}_init();
void *{prefix}_input_ptr(int index);
size_t {prefix}_input_size(int index);
size_t {prefix}_inputs();
int {prefix}_invoke();
void *{prefix}_output_ptr(int index);
size_t {prefix}_output_size(int index);
size_t {prefix}_outputs();

#endif  // {upper_prefix}_GEN_H
"""
        return code

    def generate_wrapper(
        self,
        model,
        prefix="model",
        header=True,
        legacy=False,
        debug_arena=False,
        arena_size=None,
        ops=None,
        custom_ops=None,  # TODO: implement
        registrations=None,  # TODO: implement
        ops_resolver=None,  # TODO: implement
        reporter=True,
    ):
        arena_size = arena_size if arena_size is not None else TFLMIBackend.DEFAULTS["arena_size"]
        if len(ops) > 0:

            def convert_op_name(op):
                new_name = ""
                prev = None
                for i, c in enumerate(op):
                    if c == "_":
                        prev = c
                        continue
                    elif prev is None:
                        new_name += c
                    elif (prev == "_") or prev.isdigit():
                        new_name += c
                    elif prev.isupper():
                        new_name += c.lower()
                    elif c.islower():
                        new_name += c
                    prev = c
                # workarounds for strange op names
                MAPPINGS = {
                    "Lstm": "LSTM",
                }
                for key, value in MAPPINGS.items():
                    new_name = new_name.replace(key, value)
                return new_name

            op_names = list(map(convert_op_name, ops))

            ops = op_names
        if len(custom_ops) > 0:
            raise NotImplementedError
        if len(registrations) > 0:
            raise NotImplementedError
        if ops_resolver != "mutable":
            raise NotImplementedError

        model_data = None
        with open(model, "rb") as model_buf:
            model_data = model_buf.read()

        if header:
            header_content = self.generate_header()

        wrapper_content = """
// This file is generated. Do not edit.
#include "printing.h"
#include "tensorflow/lite/micro/kernels/micro_ops.h"
#include "tensorflow/lite/micro/micro_interpreter.h"
#include "tensorflow/lite/micro/micro_mutable_op_resolver.h"
#include "tensorflow/lite/schema/schema_generated.h"
"""
        if reporter:
            wrapper_content += """#include "tensorflow/lite/micro/micro_error_reporter.h"
"""
        if legacy:
            wrapper_content += """#include "tensorflow/lite/version.h"
"""
        wrapper_content += """
#if defined __GNUC__
#define ALIGN(X) __attribute__((aligned(X)))
#elif defined _MSC_VER
#define ALIGN(X) __declspec(align(X))
#elif defined __TASKING__
#define ALIGN(X) __align(X)
#endif
"""
        if debug_arena:  # This will enable the feature only if it is not overwritten by the user/compiler
            wrapper_content += """
#ifndef DEBUG_ARENA_USAGE
#define DEBUG_ARENA_USAGE 1
#endif

"""
        wrapper_content += """const unsigned char g_model_data[] ALIGN(16) = { """
        wrapper_content += make_hex_array(model_data)
        wrapper_content += """ };

"""
        wrapper_content += self.makeCustomOpPrototypes(custom_ops)
        wrapper_content += """

namespace {
"""
        if reporter:
            wrapper_content += """
tflite::ErrorReporter *error_reporter = nullptr;
"""
        wrapper_content += """
const tflite::Model *model = nullptr;
tflite::MicroInterpreter *interpreter = nullptr;

// Create an area of memory to use for input, output, and intermediate arrays.
// Finding the minimum value for your model may require some trial and error.
constexpr int kTensorArenaSize = """
        wrapper_content += (
            str(arena_size)
            + """;
uint8_t tensor_arena[kTensorArenaSize] ALIGN(16);

class DummyReporter : public tflite::ErrorReporter {
public:
  ~DummyReporter() {}
  int Report(const char *, va_list) override { return 0; }

private:
  TF_LITE_REMOVE_VIRTUAL_DELETE
};
} // namespace

// The name of this function is important for Arduino compatibility.
"""
            + f"int {prefix}_init() {{"
            + """
  // Set up logging. Google style is to avoid globals or statics because of
  // lifetime uncertainty, but since this has a trivial destructor it's okay.
  // NOLINTNEXTLINE(runtime-global-variables)
"""
        )
        if reporter:
            wrapper_content += """
#ifdef _DEBUG
  static tflite::MicroErrorReporter micro_error_reporter;
#else
  static DummyReporter micro_error_reporter;
#endif
  error_reporter = &micro_error_reporter;
"""
        wrapper_content += """
  // Map the model into a usable data structure. This doesn't involve any
  // copying or parsing, it's a very lightweight operation.
  model = tflite::GetModel(g_model_data);
  if (model->version() != TFLITE_SCHEMA_VERSION) {
"""
        if reporter:
            wrapper_content += """
    error_reporter->Report("Model provided is schema version %d not equal "
                           "to supported version %d.",
                           model->version(), TFLITE_SCHEMA_VERSION);
"""
        wrapper_content += """
    return 1;
  }

  // This pulls in all the operation implementations we need.
  // NOLINTNEXTLINE(runtime-global-variables)
  // static tflite::ops::micro::AllOpsResolver resolver;

"""

        wrapper_content += self.make_op_registrations(ops, custom_ops, reporter=reporter)
        if reporter:
            wrapper_content += """

  // Build an interpreter to run the model with.
  static tflite::MicroInterpreter static_interpreter(
      model, resolver, tensor_arena, kTensorArenaSize, error_reporter);
"""
        else:
            wrapper_content += """

  // Build an interpreter to run the model with.
  static tflite::MicroInterpreter static_interpreter(
      model, resolver, tensor_arena, kTensorArenaSize);
"""
        wrapper_content += """
  interpreter = &static_interpreter;

  // Allocate memory from the tensor_arena for the model's tensors.
  TfLiteStatus allocate_status = interpreter->AllocateTensors();
  if (allocate_status != kTfLiteOk) {
"""
        if reporter:
            wrapper_content += """
    error_reporter->Report("AllocateTensors() failed");
"""
        wrapper_content += """
    return 1;
  }
  return 0;
}
"""

        wrapper_content += f"""
void *{prefix}_input_ptr(int id) {{
  return interpreter->input(id)->data.data;
}}

size_t {prefix}_input_size(int id) {{
  return interpreter->input(id)->bytes;
}}

size_t {prefix}_inputs() {{
  return interpreter->inputs_size();
}}

void *{prefix}_output_ptr(int id) {{
  return interpreter->output(id)->data.data;
}}

size_t {prefix}_output_size(int id) {{
  return interpreter->output(id)->bytes;
}}

size_t {prefix}_outputs() {{
  return interpreter->outputs_size();
}}

int {prefix}_invoke() {{
  // Run inference, and report any error
  TfLiteStatus invoke_status = interpreter->Invoke();
  if (invoke_status != kTfLiteOk) {{
"""
        if reporter:
            # TODO
            wrapper_content += """
    error_reporter->Report("Invoke failed\\n");
"""
        wrapper_content += """
    return 1;
  }
#if DEBUG_ARENA_USAGE
  size_t used = interpreter->arena_used_bytes();
"""
        if reporter:
            # TODO
            wrapper_content += """
  error_reporter->Report("Arena Usage after model invocation: %d bytes\\n", used);
"""
        else:
            wrapper_content += """
  // printf("Arena Usage after model invocation: %d bytes\\n", used);
  // printf("# Arena Usage: %d\\n", used);
  mlonmcu_printf("# Arena Usage: %d\\n", used);
  // MicroPrintf("# Arena Usage: %d\\n", used);

"""
        wrapper_content += """
#endif  // DEBUG_ARENA_USAGE
  return 0;
}
"""
        if header:
            return wrapper_content, header_content
        else:
            return wrapper_content


class TFLMIBackend(TFLMBackend):
    name = "tflmi"

    FEATURES = TFLMBackend.FEATURES | {"debug_arena"}

    DEFAULTS = {
        **TFLMBackend.DEFAULTS,
        "arena_size": 2**20,  # 1 MB
        "debug_arena": False,
        "ops": [],
        "custom_ops": [],
        "registrations": {},
        "ops_resolver": "mutable",
        "legacy": False,
        "reporter": False,  # Has to be disabled for support with latest upstream
    }

    def __init__(self, features=None, config=None):
        super().__init__(features=features, config=config)
        self.codegen = TFLMICodegen()
        self.model_data = None
        self.prefix = "model"  # Without the _
        self.artifacts = (
            []
        )  # TODO: either make sure that ony one model is processed at a time or move the artifacts to the methods
        # TODO: decide if artifacts should be handled by code (str) or file path or binary data

    @property
    def legacy(self):
        value = self.config["legacy"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def debug_arena(self):
        value = self.config["debug_arena"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def arena_size(self):
        return int(self.config["arena_size"])

    @property
    def ops(self):
        value = self.config["ops"]
        return str2list(value) if isinstance(value, str) else value

    @property
    def custom_ops(self):
        value = self.config["custom_ops"]
        return str2list(value) if isinstance(value, str) else value

    @property
    def registrations(self):
        value = self.config["registrations"]
        return str2list(value) if isinstance(value, str) else value

    @property
    def ops_resolver(self):
        return self.config["ops_resolver"]

    @property
    def reporter(self):
        value = self.config["reporter"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def generate(self) -> Tuple[dict, dict]:
        artifacts = []
        assert self.model is not None
        wrapper_code, header_code = self.codegen.generate_wrapper(
            self.model,
            prefix=self.prefix,
            header=True,
            arena_size=self.arena_size,
            debug_arena=self.debug_arena,
            ops=self.ops,
            custom_ops=self.custom_ops,
            registrations=self.registrations,
            ops_resolver=self.ops_resolver,
            legacy=self.legacy,
            reporter=self.reporter,
        )
        artifacts.append(Artifact(f"{self.prefix}.cc", content=wrapper_code, fmt=ArtifactFormat.SOURCE))
        artifacts.append(
            Artifact(
                f"{self.prefix}.cc.h",
                content=header_code,
                fmt=ArtifactFormat.SOURCE,
                optional=False,
            )
        )
        workspace_size_artifact = Artifact(
            "tflmi_arena_size.txt", content=f"{self.arena_size}", fmt=ArtifactFormat.TEXT
        )
        artifacts.append(workspace_size_artifact)
        # TODO: stdout_artifact (Would need to invoke TFLMI in subprocess to get stdout)
        return {"default": artifacts}, {}


if __name__ == "__main__":
    sys.exit(
        main(
            TFLMIBackend,
            args=sys.argv[1:],
        )
    )  # pragma: no cover
# TODO: pin defaults to backend class? -> how to defien features and config?



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions of a task registry used to automatically install dependencies."""

from functools import wraps
import itertools
from enum import Enum
import time
from typing import List, Tuple
import networkx as nx
from networkx.drawing.nx_agraph import write_dot
from tqdm import tqdm

from mlonmcu.logging import get_logger

logger = get_logger()


def get_combs(data) -> List[dict]:
    """Utility which returns combinations of the input data.

    Parameters
    ----------
    data : dict
        Input dictionary

    Returns
    -------
    combs : list
        All combinations of the input data.

    Examples
    --------
    >>> get_combs({"foo": [False, True], "bar": [5, 10]})
    [{"foo": False, "bar": 5}, {"foo": False, "bar": 10}, {"foo": True, "bar": 5}, {"foo": True, "bar": 10}]
    """
    keys = list(data.keys())
    values = list(data.values())
    prod = list(itertools.product(*values))
    if len(prod) == 1:
        if len(prod[0]) == 0:
            prod = []
    combs = [dict(zip(keys, p)) for p in prod]
    return combs


class TaskType(Enum):
    """Enumeration for the task type."""

    MISC = 0
    FRAMEWORK = 1
    BACKEND = 2
    TOOLCHAIN = 3
    TARGET = 4
    FRONTEND = 5
    OPT = 6
    FEATURE = 7
    PLATFORM = 8


class TaskGraph:
    """Task graph object.

    Attributes
    ----------
    names : list
        list of task names in the graph
    dependencies : dict
        Dependencies between task artifacts
    providers : dict
        Providers for all the artifacts

    Examples
    -------
    TODO
    """

    def __init__(self, names: List[str], dependencies: dict, providers: dict):
        self.names = names
        self.dependencies = dependencies
        self.providers = providers

    def get_graph(self) -> Tuple[list, list]:
        """Get nodes and edges of the task graph.

        Returns
        -------
        nodes : list
            List of edges
        edges : list
            List of edge tuples.
        """
        nodes = list(self.names)
        edges = []
        for dest, deps in self.dependencies.items():
            for dep in deps:
                if dep not in self.providers.keys():
                    raise RuntimeError(f"Unable to resolve dependency '{dep}'")
                src = self.providers[dep]
                edge = (src, dest)
                edges.append(edge)
        # Remove duplicates
        edges = list(dict.fromkeys(edges))
        return nodes, edges

    def get_order(self) -> list:
        """Get execution order of tasks via topological sorting."""
        nodes, edges = self.get_graph()
        graph = nx.DiGraph(edges)
        graph.add_nodes_from(nodes)
        order = list(nx.topological_sort(graph))
        return order

    def export_dot(self, path):
        """Visualize the task dependency graph."""
        nodes, edges = self.get_graph()
        graph = nx.DiGraph(edges)
        graph.add_nodes_from(nodes)
        # order = list(nx.topological_sort(graph))
        # TODO: annotate with order
        # TODO: also export order as extra graph
        write_dot(graph, path)


class TaskFactory:
    """Class which is used to register all available tasks and their annotations.

    Attributes
    ----------
    registry : dict
        Mapping of task names and their actual function
    dependencies : dict
        Mapping of task dependencies
    providers : dict
        Mapping of which task provides which artifacts
    types : dict
        Mapping of task types
    validates : dict
        Mapping of validation functions for the tasks
    changed : list
        List of tasks?artifacts which have changed recently
    """

    def __init__(self):
        self.registry = {}
        self.dependencies = {}
        self.providers = {}
        self.types = {}
        self.params = {}
        self.validates = {}
        self.changed = []  # Main problem: per

    def reset_changes(self):
        """Reset all pending changes."""
        self.changed = []

    def needs(self, keys, force=True):
        """Decorator which registers the artifacts a task needs to be processed."""

        def real_decorator(function):
            name = function.__name__
            if name in self.dependencies:
                self.dependencies[name].extend(keys)
            else:
                self.dependencies[name] = keys

            @wraps(function)
            def wrapper(*args, **kwargs):
                # logger.debug("Checking inputs...")
                if force:
                    context = args[0]
                    variables = context.cache._vars
                    for key in keys:
                        if key not in variables.keys() or variables[key] is None:
                            raise RuntimeError(f"Task '{name}' needs the value of '{key}' which is not set")
                retval = function(*args, **kwargs)
                return retval

            return wrapper

        return real_decorator

    def optional(self, keys):
        """Decorator for optional task requirements."""
        return self.needs(keys, force=False)

    def removes(self, keys):
        """Decorator for cleanuo tasks."""

        # TODO: implementation
        def real_decorator(function):
            @wraps(function)
            def wrapper(*args, **kwargs):
                retval = function(*args, **kwargs)
                return retval

            return wrapper

        return real_decorator

    # def optional(self, keys):
    #     def real_decorator(function):
    #         name = function.__name__
    #         if name in self.dependencies:
    #             self.dependencies[name].extend(keys)
    #         else:
    #             self.dependencies[name] = keys
    #         @wraps(function)
    #         def wrapper(*args, **kwargs):
    #             retval = function(*args, **kwargs)
    #             return retval
    #         return wrapper
    #     return real_decorator

    def provides(self, keys):
        """Decorator which registers what a task provides."""

        def real_decorator(function):
            name = function.__name__
            for key in keys:
                self.providers[key] = name

            @wraps(function)
            def wrapper(*args, **kwargs):
                context = args[0]
                for key in keys:
                    if key in context.cache._vars:
                        del context.cache._vars[key]  # Unset the value before calling function
                retval = function(*args, **kwargs)
                if retval is not False:
                    # logger.debug("Checking outputs...")
                    variables = context.cache._vars
                    for key in keys:
                        if key not in variables.keys() or variables[key] is None:
                            raise RuntimeError(f"Task '{name}' did not set the value of '{key}'")
                return retval

            self.registry[name] = wrapper
            return wrapper

        return real_decorator

    def param(self, flag, options):
        """Decorator which registers available task parameters."""
        if not isinstance(options, list):
            options = [options]

        def real_decorator(function):
            name = function.__name__
            if name in self.params:
                self.params[name][flag] = options
            else:
                self.params[name] = {flag: options}

            @wraps(function)
            def wrapper(*args, **kwargs):
                retval = function(*args, **kwargs)
                return retval

            return wrapper

        return real_decorator

    def validate(self, func):
        """Decorator which registers validation functions for a task."""

        def real_decorator(function):
            name = function.__name__
            self.validates[name] = func
            return function

        return real_decorator

    def register(self, category=TaskType.MISC):
        """Decorator which actually registers a task in the registry."""

        def real_decorator(function):
            name = function.__name__

            @wraps(function)
            def wrapper(*args, rebuild=False, progress=False, **kwargs):
                combs = get_combs(self.params[name])

                def get_valid_combs(combs):
                    ret = []
                    for comb in combs:
                        if name in self.validates:
                            check = self.validates[name](args[0], params=comb)
                            if not check:
                                continue
                        ret.append(comb)
                    return ret

                combs_ = get_valid_combs(combs)

                def process(name_, params=None, rebuild=False):
                    if not params:
                        params = []
                    rebuild = rebuild
                    if name in self.dependencies:
                        for dep in self.dependencies[name]:
                            if dep in self.changed:
                                rebuild = True
                                break
                    retval = function(*args, params=params, rebuild=rebuild, **kwargs)
                    if retval:
                        keys = [key for key, provider in self.providers.items() if provider == name]
                        for key in keys:
                            if key not in self.changed:
                                self.changed.append(key)
                    # logger.debug("Processed task:", function.__name__)
                    return retval

                if progress:
                    pbar = tqdm(
                        total=max(len(combs_), 1),
                        desc="Processing",
                        ncols=100,
                        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}s]",
                        leave=None,
                    )
                else:
                    pbar = None
                if len(combs_) == 0:
                    if pbar:
                        pbar.set_description(f"Processing: {name}")
                    else:
                        logger.info("Processing task: %s", name)
                    time.sleep(0.1)
                    check = True
                    if len(combs) > 0:
                        check = False
                    else:
                        if name in self.validates:
                            check = self.validates[name](args[0], params={})
                    if check:
                        start = time.time()
                        retval = process(name, rebuild=rebuild)
                        end = time.time()
                        diff = end - start
                        minutes = int(diff // 60)
                        seconds = int(diff % 60)
                        duration_str = f"{seconds}s" if minutes == 0 else f"{minutes}m{seconds}s"
                        if not pbar:
                            logger.debug("-> Done (%s)", duration_str)
                        # TODO: move this to helper func
                    else:
                        logger.debug("-> Skipped")
                        retval = False
                    if pbar:
                        pbar.update(1)
                else:
                    for comb in combs_:  # TODO process in parallel?
                        extended_name = name + str(comb)
                        if pbar:
                            pbar.set_description(f"Processing - {extended_name}")
                        else:
                            logger.info("Processing task: %s", extended_name)
                        time.sleep(0.1)
                        start = time.time()
                        retval = process(extended_name, params=comb, rebuild=rebuild)
                        end = time.time()
                        diff = end - start
                        minutes = int(diff // 60)
                        seconds = int(diff % 60)
                        duration_str = f"{seconds}s" if minutes == 0 else f"{minutes}m{seconds}s"
                        if not pbar:
                            logger.debug("-> Done (%s)", duration_str)
                            # TODO: move this to helper func
                        else:
                            pbar.update(1)
                if pbar:
                    pbar.close()
                return retval

            self.registry[name] = wrapper
            self.types[name] = category
            self.params[name] = {}
            return wrapper

        return real_decorator



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import time
import tempfile
import multiprocessing
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Tuple, List

from mlonmcu.feature.features import get_matching_features
from mlonmcu.models.model import (
    ModelFormats,
    Model,
    ExampleProgram,
    EmbenchProgram,
    TaclebenchProgram,
    PolybenchProgram,
    CoremarkProgram,
    DhrystoneProgram,
    MathisProgram,
    MibenchProgram,
)
from mlonmcu.models.lookup import lookup_models
from mlonmcu.feature.type import FeatureType
from mlonmcu.config import filter_config, str2bool
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.setup import utils
from mlonmcu.target.metrics import Metrics

from mlonmcu.logging import get_logger


logger = get_logger()


class Frontend(ABC):
    FEATURES = {"validate"}

    DEFAULTS = {
        "use_inout_data": False,
        # TODO: print_outputs for frontends
    }

    REQUIRED = set()
    OPTIONAL = set()

    def __init__(self, name, input_formats=None, output_formats=None, features=None, config=None):
        self.name = name
        self.input_formats = input_formats if input_formats else []
        self.output_formats = output_formats if output_formats else []
        self.config = config if config else {}
        self.features = self.process_features(features)
        self.config = filter_config(self.config, self.name, self.DEFAULTS, self.OPTIONAL, self.REQUIRED)

    def __repr__(self):
        probs = []
        if self.name:
            probs.append(self.name)
        if self.features and len(self.features) > 0:
            probs.append(str(self.features))
        if self.config and len(self.config) > 0:
            probs.append(str(self.config))
        return "Frontend(" + ",".join(probs) + ")"

    @property
    def use_inout_data(self):
        value = self.config["use_inout_data"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def supports_formats(self, ins=None, outs=None):
        """Returs true if the frontend can handle at least one combination of input and output formats."""
        assert ins is not None or outs is not None, "Please provide a list of input formats, outputs formats or both"
        ret = True
        if ins:
            if not isinstance(ins, list):
                ins = [ins]
            supported = any(fmt in self.input_formats for fmt in ins)
            ret = ret and supported
        if outs:
            if not isinstance(outs, list):
                outs = [outs]
            supported = any(fmt in self.output_formats for fmt in ins)
            ret = ret and supported
        return ret

    def lookup_models(self, names, config=None, context=None):
        return lookup_models(names, frontends=[self], config=config, context=context)

    def process_features(self, features):
        if features is None:
            return []
        features = get_matching_features(features, FeatureType.FRONTEND)
        for feature in features:
            assert (  # If this assertion occurs, continue with the next frontend instead of failing
                # (TODO: create custom exception type)
                feature.name
                in self.FEATURES
            ), f"Incompatible feature: {feature.name}"
            # Instead we might introduce self.compatible and set it to true at this line
            feature.used = True
            feature.add_frontend_config(self.name, self.config)
            feature.update_formats(self.name, self.input_formats, self.output_formats)
        return features

    @abstractmethod
    # def produce_artifacts(self, model):
    def produce_artifacts(self, model):
        pass

    def process_metadata(self, model, cfg=None):
        model_dir = Path(model.paths[0]).parent.resolve()
        metadata = model.metadata
        in_paths = []
        out_paths = []
        input_shapes = {}
        output_shapes = {}
        input_types = {}
        output_types = {}
        if metadata is not None and "network_parameters" in metadata:
            network = metadata["network_parameters"]
            assert "input_nodes" in network
            ins = network["input_nodes"]
            for inp in ins:
                name = inp.get("name", None)
                shape = inp.get("shape", None)
                ty = inp.get("type", None)
                if name and shape:
                    input_shapes[name] = shape
                if name and ty:
                    input_types[name] = ty
                if self.use_inout_data:
                    if "example_input" in inp and "path" in inp["example_input"]:
                        in_data_dir = Path(inp["example_input"]["path"])
                        # TODO: this will only work with relative paths to model dir! (Fallback to parent directories?)
                        in_path = model_dir / in_data_dir
                        assert (
                            in_path.is_dir()
                        ), f"Input data directory defined in model metadata does not exist: {in_path}"
                        in_paths.append(in_path)
            assert "output_nodes" in network
            outs = network["output_nodes"]
            for outp in outs:
                name = outp.get("name", None)
                shape = outp.get("shape", None)
                ty = outp.get("type", None)
                if name and shape:
                    output_shapes[name] = shape
                if name and ty:
                    output_types[name] = ty
                if self.use_inout_data:
                    if "test_output_path" in outp:
                        out_data_dir = Path(outp["test_output_path"])
                        out_path = model_dir / out_data_dir
                        assert (
                            in_path.is_dir()
                        ), f"Output data directory defined in model metadata does not exist: {out_path}"
                        out_paths.append(out_path)
        else:
            fallback_in_path = model_dir / "input"
            if fallback_in_path.is_dir():
                in_paths.append(fallback_in_path)
            fallback_out_path = model_dir / "output"
            if fallback_out_path.is_dir():
                out_paths.append(fallback_out_path)

        if metadata is not None and "backends" in metadata:
            assert cfg is not None
            backend_options = metadata["backends"]
            for backend in backend_options:
                if backend_options[backend] is not None:
                    flattened = {f"{backend}.{key}": value for key, value in backend_options[backend].items()}
                    cfg.update(flattened)

        # Detect model support code (Allow overwrite in metadata YAML)
        support_path = model_dir / "support"
        if support_path.is_dir():
            assert cfg is not None
            # TODO: onlu overwrite if unset?
            cfg.update({"mlif.model_support_dir": support_path})
            # cfg.update({"espidf.model_support_dir": support_path})
            # cfg.update({"zephyr.model_support_dir": support_path})
        if len(in_paths) > 0:
            cfg.update({"mlif.input_data_path": in_paths})
            # cfg.update({"espidf.input_data_path": in_paths})
            # cfg.update({"zephyr.input_data_path": in_paths})
        if len(out_paths) > 0:
            cfg.update({"mlif.output_data_path": out_paths})
            # cfg.update({"espidf.output_data_path": out_paths})
            # cfg.update({"zephyr.output_data_path": out_paths})
        if len(input_shapes) > 0:
            cfg.update({f"{model.name}.input_shapes": input_shapes})
        if len(output_shapes) > 0:
            cfg.update({f"{model.name}.output_shapes": output_shapes})
        if len(input_types) > 0:
            cfg.update({f"{model.name}.input_types": input_types})
        if len(output_types) > 0:
            cfg.update({f"{model.name}.output_types": output_types})

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = []

        count = len(model.paths)
        assert count == len(model.formats)
        assert count > 0, f"'{self.name}' frontend expects at least one model"
        max_ins = len(self.input_formats)
        assert count <= max_ins, f"'{self.name}' frontend did not expect more than {max_ins} models"
        formats = model.formats
        assert self.supports_formats(formats), f"Invalid model format for '{self.name}' frontend"

        artifacts = self.produce_artifacts(model)
        if not isinstance(artifacts, list):
            artifacts = [artifacts]
        assert len(artifacts) > 0, f"'{self.name}' frontend should produce at least one model"
        max_outs = len(self.output_formats)
        assert len(artifacts) <= max_outs, f"'{self.name}' frontend should not return more than {max_outs}"

        # If we want to use the same instance of this Frontend in parallel, we need to get rid of self.artifacts...
        return {"default": artifacts}, {}

    def generate_artifacts(self, model) -> List[Artifact]:
        start_time = time.time()
        artifacts, metrics = self.generate(model)
        # TODO: do something with out?
        end_time = time.time()
        diff = end_time - start_time
        if len(metrics) == 0:
            metrics = {"default": Metrics()}
        for name, metrics_ in metrics.items():
            if name == "default":
                metrics_.add("Load Stage Time [s]", diff, True)
            content = metrics_.to_csv(include_optional=True)
            artifact = Artifact("load_metrics.csv", content=content, fmt=ArtifactFormat.TEXT, flags=["metrics"])
            if name not in artifacts:
                artifacts[name] = []
            artifacts[name].append(artifact)
        self.artifacts = artifacts
        return artifacts

    def export_artifacts(self, path):
        assert len(self.artifacts) > 0, "No artifacts found, please run generate_artifacts() first"

        if not isinstance(path, Path):
            path = Path(path)

        is_dir = len(path.suffix) == 0
        if is_dir:
            assert (
                path.is_dir()
            ), "The supplied path does not exists."  # Make sure it actually exists (we do not create it by default)
            for artifact in self.artifacts:
                artifact.export(path)
        else:
            raise NotImplementedError

    def get_platform_config(self, platform):
        return {}

    def add_platform_config(self, platform, config):
        config.update(self.get_platform_config(platform))

    def get_platform_defs(self, platform):
        return {}

    def add_platform_defs(self, platform, defs):
        defs.update(self.get_platform_defs(platform))


class SimpleFrontend(Frontend):
    """An abstract frontend with equivalent input and output formats."""

    # Assumption: only raw model data
    def __init__(self, name, fmt, features=None, config=None):
        super().__init__(
            name,
            input_formats=[fmt],
            output_formats=[fmt],
            features=features,
            config=config,
        )

    def produce_artifacts(self, model):
        assert len(self.input_formats) == len(self.output_formats) == len(model.paths) == 1
        artifacts = []
        name = model.name
        path = model.paths[0]
        ext = self.input_formats[0].extension
        with open(path, "rb") as handle:  # TODO: is an onnx model raw data or text?
            raw = handle.read()
            artifacts.append(Artifact(f"{name}.{ext}", raw=raw, fmt=ArtifactFormat.RAW, flags=["model"]))
        return artifacts


# TODO: move to frontends.py
# TODO: frontend parsed metadata instead of lookup.py?
# TODO: how to find inout_data?
class TfLiteFrontend(SimpleFrontend):
    FEATURES = Frontend.FEATURES | {"visualize", "split_layers", "tflite_analyze"}

    DEFAULTS = {
        **Frontend.DEFAULTS,
        "visualize_enable": False,
        "visualize_script": None,
        "split_layers": False,
        "pack_script": None,
        "analyze_enable": False,
        "analyze_script": None,
    }

    REQUIRED = Frontend.REQUIRED

    OPTIONAL = SimpleFrontend.OPTIONAL | {"tflite_analyze.script"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "tflite",
            ModelFormats.TFLITE,
            features=features,
            config=config,
        )

    @property
    def visualize_enable(self):
        value = self.config["visualize_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def split_layers(self):
        value = self.config["split_layers"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def visualize_script(self):
        return self.config["visualize_script"]

    @property
    def pack_script(self):
        return self.config["pack_script"]

    @property
    def analyze_enable(self):
        value = self.config["analyze_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def analyze_script(self):
        return self.config["analyze_script"]

    def produce_artifacts(self, model):
        assert len(self.input_formats) == len(model.paths) == 1
        artifacts = []

        name = model.name
        # assert "/" not in name
        if "/" in name:
            name = name.rsplit("/", 1)[-1]
        path = model.paths[0]
        ext = self.input_formats[0].extension
        with open(path, "rb") as handle:
            raw = handle.read()
            artifacts.append(Artifact(f"{name}.{ext}", raw=raw, fmt=ArtifactFormat.RAW, flags=["model"]))

        if self.analyze_enable:
            with tempfile.TemporaryDirectory() as tmpdirname:
                out_file = str(Path(tmpdirname) / "tflite_analyze.csv")

                args = [
                    path,
                    "--csv",
                    out_file,
                    "--ops",
                    "--estimate-macs",
                    "--estimate-rom",
                    "--estimate-ram",
                ]

                assert self.analyze_script is not None
                assert Path(self.analyze_script).is_file(), f"Script {self.analyze_script} not found."
                utils.python(self.analyze_script, *args)

                with open(out_file, "r") as handle:
                    tflite_analyze_csv = handle.read()

                tflite_analyze_artifact = Artifact(
                    "tflite_analyze.csv",
                    content=tflite_analyze_csv,
                    fmt=ArtifactFormat.TEXT,
                )
                artifacts.append(tflite_analyze_artifact)
        if self.visualize_enable:
            assert self.visualize_script is not None

            in_file = model.paths[0]
            ext = "html"
            with tempfile.TemporaryDirectory() as tmpdirname:
                out_file = str(Path(tmpdirname) / f"tflite_visualize.{ext}")

                utils.python(self.visualize_script, in_file, out_file)

                with open(out_file, "r") as handle:
                    tflite_visualize_text = handle.read()

                tflite_visualize_artifact = Artifact(
                    f"tflite_visualize.{ext}",
                    content=tflite_visualize_text,
                    fmt=ArtifactFormat.TEXT,
                )
                artifacts.append(tflite_visualize_artifact)

        if self.visualize_enable and self.analyze_enable:
            assert len(self.output_formats) == 3
        elif self.visualize_enable or self.analyze_enable:
            assert len(self.output_formats) == 2
        else:
            assert len(self.output_formats) == 1

        return artifacts

    def generate(self, model) -> Tuple[dict, dict]:
        if self.split_layers:
            artifacts = {}

            name = model.name
            path = model.paths[0]
            formats = model.formats
            assert self.supports_formats(formats), f"Invalid model format for '{self.name}' frontend"

            ret = self.produce_artifacts(model)
            if not isinstance(ret, list):
                ret = [ret]
            assert len(ret) > 0, f"'{self.name}' frontend should produce at least one model"
            max_outs = len(self.output_formats)
            assert len(ret) <= max_outs, f"'{self.name}' frontend should not return more than {max_outs}"
            artifacts["default"] = ret
            with tempfile.TemporaryDirectory() as tmpdirname:

                def get_num_layers(file):
                    tflite_pack_args = [path, "--count-layers", "--noop"]
                    out = utils.exec_getout(self.pack_script, *tflite_pack_args)
                    matches = re.compile(r"Found\s(\d+)\slayers.").findall(out)
                    assert len(matches) == 1
                    num = int(matches[0])
                    return num

                replace = False
                # replace = True
                drop = False

                # drop = True
                def gen_layer_files(file, dest):
                    results = []
                    num_layers = get_num_layers(file)
                    assert num_layers > 0
                    keep = None
                    if replace:
                        assert keep is not None and len(keep) == 1
                    for i in range(num_layers):
                        if keep and i not in keep:
                            continue
                        out_name = f"layer{i}"
                        out_file = Path(dest) / out_name
                        tflite_pack_args = [path, "-k", str(i), "--out", out_file]
                        utils.exec_getout(self.pack_script, *tflite_pack_args)
                        assert out_file.is_file()
                        results.append(out_file)
                    return results

                layer_files = gen_layer_files(path, tmpdirname)

                for i, layer_file in enumerate(layer_files):
                    subrun = f"layer{i}"
                    layer_name = f"{name}_{subrun}"
                    layer_model = Model(layer_name, [layer_file])
                    ret = self.produce_artifacts(layer_model)
                    if not isinstance(ret, list):
                        ret = [ret]
                    assert len(ret) > 0, f"'{self.name}' frontend should produce at least one model"
                    max_outs = len(self.output_formats)
                    assert len(ret) <= max_outs, f"'{self.name}' frontend should not return more than {max_outs}"
                    if replace:
                        subrun = "default"
                    artifacts[subrun] = ret
                if drop:
                    del artifacts["default"]

            return artifacts, {}
        else:
            return super().generate(model)


class RelayFrontend(SimpleFrontend):
    FEATURES = Frontend.FEATURES | {"relayviz"}

    DEFAULTS = {**Frontend.DEFAULTS, "visualize_graph": False, "relayviz_plotter": "term"}

    REQUIRED = Frontend.REQUIRED | {"tvm.build_dir", "tvm.pythonpath"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "relay",
            ModelFormats.RELAY,
            features=features,
            config=config,
        )

    @property
    def visualize_graph(self):
        value = self.config["visualize_graph"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def relayviz_plotter(self):
        return self.config["relayviz_plotter"]

    @property
    def tvm_build_dir(self):
        return self.config["tvm.build_dir"]

    @property
    def tvm_pythonpath(self):
        return self.config["tvm.pythonpath"]

    def produce_artifacts(self, model):
        assert len(self.input_formats) == len(model.paths) == 1
        artifacts = []

        name = model.name
        path = model.paths[0]
        ext = self.input_formats[0].extension
        with open(path, "rb") as handle:  # TODO: is an onnx model raw data or text?
            raw = handle.read()
            artifacts.append(Artifact(f"{name}.{ext}", raw=raw, fmt=ArtifactFormat.RAW, flags=["model"]))

        if not self.visualize_graph:
            assert len(self.output_formats) == 1
        else:
            assert len(self.output_formats) == 2

            def _relayviz(in_file, out_file, plotter_name, env={}):
                import sys
                import os

                sys.path.append(env["PYTHONPATH"])
                os.environ["TVM_LIBRARY_PATH"] = env["TVM_LIBRARY_PATH"]
                from tvm import parser
                from tvm.contrib import relay_viz
                from tvm.contrib.relay_viz.terminal import TermPlotter
                from tvm.contrib.relay_viz.dot import DotPlotter

                if plotter_name == "term":
                    plotter_cls = TermPlotter
                elif plotter_name == "dot":
                    plotter_cls = DotPlotter
                else:
                    raise RuntimeError(f"Invalid plotter name: {plotter_name}")

                with open(in_file, "r", encoding="utf-8") as relay_text:
                    text = relay_text.read()

                mod = parser.fromtext(text)

                plotter_inst = plotter_cls()
                viz = relay_viz.RelayVisualizer(mod, plotter=plotter_inst)
                out_file_base = os.path.splitext(out_file)[0]
                viz.render(filename=out_file_base)

            in_file = model.paths[0]
            ext = "" if self.relayviz_plotter == "term" else "pdf"
            with tempfile.TemporaryDirectory() as tmpdirname:
                out_file = str(Path(tmpdirname) / (f"relayviz.{ext}" if len(ext) > 0 else "relayviz"))
                proc = multiprocessing.Process(
                    target=_relayviz,
                    args=[in_file, out_file, self.relayviz_plotter],
                    kwargs={"env": {"PYTHONPATH": self.tvm_pythonpath, "TVM_LIBRARY_PATH": self.tvm_build_dir}},
                )
                proc.start()
                proc.join()

                if self.relayviz_plotter == "term":
                    with open(out_file, "r") as handle:
                        relayviz_text = handle.read()

                    relayviz_artifact = Artifact(
                        "relayviz.txt",
                        content=relayviz_text,
                        fmt=ArtifactFormat.TEXT,
                    )
                else:
                    with open(out_file, "rb") as handle:
                        relayviz_data = handle.read()

                    relayviz_artifact = Artifact(
                        f"relayviz.{ext}",
                        raw=relayviz_data,
                        fmt=ArtifactFormat.RAW,
                    )
                artifacts.append(relayviz_artifact)

        return artifacts


class PackedFrontend(Frontend):  # Inherit from TFLiteFrontend? -> how to do constructor?
    FEATURES = Frontend.FEATURES | {"packing", "packed"}

    DEFAULTS = {
        **Frontend.DEFAULTS,
        "ignore_existing": True,
        "fake_pack": False,  # Pretend that every compatible tensor is packable
        # (best case scenerio, TODO: rename to force_pack?)
        "use_packed": True,
        "check": False,  # Unimplemented
    }

    REQUIRED = {"packer.exe"}  # TODO move to feature?

    def __init__(self, features=None, config=None):
        super().__init__(name="packed", features=features, config=config)
        if self.fake_pack or self.ignore_existing:
            # assert self.use_packed
            self.input_formats = [ModelFormats.TFLITE]
        else:
            self.input_formats = [ModelFormats.PACKED, ModelFormats.TFLITE]

        # if self.use_packed:
        self.output_formats = [
            ModelFormats.PACKED,
            ModelFormats.TFLITE,
        ]  # Always copy over the input model as intermediate artifact for reproducability
        # TODO: add a Frontend.DEFAULT which can be used to turn off this behavoir (keep intermediate?)
        # else:
        #     self.output_formats = [ModelFormats.TFLITE]
        # Order of formats ir irrelevant here, hot for artifacts, the first one will always be the main object

    @property
    def ignore_existing(self):
        value = self.config["ignore_existing"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def fake_pack(self):
        value = self.config["fake_pack"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def use_packed(self):
        value = self.config["use_packed"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def check(self):
        value = self.config["check"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def produce_artifacts(self, model):
        tflite_data = None
        packed_data = None
        name = model.name

        if self.fake_pack or self.ignore_existing:  # -> ins: TFLITE
            # assert self.use_packed
            tflite_path = model.paths[0]

            with open(tflite_path, "rb") as handle:
                tflite_data = handle.read()
        else:  # -> ins: TFLITE, PACKED
            for path, fmt in zip(model.paths, model.formats):
                data_in = None
                with open(path, "rb") as handle:
                    data_in = handle.read()
                if fmt == ModelFormats.PACKED:
                    packed_data = data_in
                    break
                elif fmt == ModelFormats.TFLITE:
                    # tflite_data_in = data_in
                    break
                else:
                    raise RuntimeError(f"Unexpected model format: {fmt}")

        if packed_data is None:
            # Do packing
            with tempfile.TemporaryDirectory() as tmpdirname:
                logger.debug("Using temporary directory for packing results: %s", tmpdirname)
                packer_exe = self.config["packer_exe"]
                assert packer_exe is not None
                in_file = Path(tmpdirname) / "in.tflite"
                with open(in_file, "wb") as handle:
                    handle.write(tflite_data)
                out_file = Path(tmpdirname) / "out.tflm"
                utils.exec_getout(packer_exe, in_file, out_file)
                with open(out_file, "rb") as handle:
                    packed_data = handle.read()

        if self.check:
            raise NotImplementedError

        tflite_artifact = Artifact(
            f"{name}.tflite",
            raw=tflite_data,
            fmt=ArtifactFormat.RAW,
            flags=["model"],
            optional=self.use_packed,
        )
        packed_artifact = Artifact(
            f"{name}.tflm",
            raw=packed_data,
            fmt=ArtifactFormat.RAW,
            optional=not self.use_packed,
            flags=["model"],
        )

        if self.use_packed:
            return [packed_artifact, tflite_artifact]
        else:
            return [tflite_artifact, packed_artifact]


class ONNXFrontend(SimpleFrontend):
    def __init__(self, features=None, config=None):
        super().__init__(
            "onnx",
            ModelFormats.ONNX,
            features=features,
            config=config,
        )


class PBFrontend(SimpleFrontend):
    def __init__(self, features=None, config=None):
        super().__init__(
            "pb",
            ModelFormats.PB,
            features=features,
            config=config,
        )


class PaddleFrontend(SimpleFrontend):
    def __init__(self, features=None, config=None):
        super().__init__(
            "paddle",
            ModelFormats.PADDLE,
            features=features,
            config=config,
        )


class ExampleFrontend(SimpleFrontend):
    def __init__(self, features=None, config=None):
        super().__init__(
            "example",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        return ["hello_world", "foobar"]

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("example/", "")
            if name in self.supported_names:
                hint = ExampleProgram(
                    name,
                    alt=f"example/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "example"
        return ret


class EmbenchFrontend(SimpleFrontend):
    REQUIRED = {"embench.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "embench",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        # TODO: automatic lookup
        return [
            "edn",
            "md5sum",
            "nettle-sha256",
            "nettle-aes",
            "ud",
            "matmult-int",
            "aha-mont64",
            "huffbench",
            "cubic",
            "nbody",
            "sglib-combined",
            "crc32",
            "wikisort",
            "slre",
            "qrduino",
            "minver",
            "picojpeg",
            "tarfind",
            "st",
            "nsichneu",
            "statemate",
            "primecount",
        ]

    # @property
    # def skip_backend(self):
    #     return True

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("embench/", "")
            if name in self.supported_names:
                hint = EmbenchProgram(
                    name,
                    alt=f"embench/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["EMBENCH_DIR"] = Path(self.config["embench.src_dir"])
        return ret

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "embench"
        return ret


class TaclebenchFrontend(SimpleFrontend):
    REQUIRED = {"taclebench.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "taclebench",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        # TODO: automatic lookup
        return [
            "test/test3",
            "test/cover",
            "test/duff",
            "app/powerwindow",
            "app/lift",
            "kernel/deg2rad",
            "kernel/matrix1",
            "kernel/binarysearch",
            "kernel/pm",
            "kernel/sha",
            "kernel/filterbank",
            "kernel/md5",
            "kernel/fir2dim",
            "kernel/fft",
            "kernel/minver",
            "kernel/lms",
            "kernel/bitcount",
            "kernel/st",
            "kernel/bsort",
            "kernel/bitonic",
            "kernel/iir",
            "kernel/prime",
            "kernel/jfdctint",
            "kernel/recursion",
            "kernel/complex_updates",
            "kernel/cosf",
            "kernel/insertsort",
            "kernel/fac",
            "kernel/rad2deg",
            "kernel/isqrt",
            "kernel/cubic",
            "kernel/ludcmp",
            "kernel/quicksort",
            "kernel/countnegative",
            "sequential/epic",
            "sequential/huff_dec",
            "sequential/fmref",
            "sequential/h264_dec",
            "sequential/dijkstra",
            "sequential/adpcm_dec",
            "sequential/adpcm_enc",
            "sequential/gsm_dec",
            "sequential/rijndael_dec",
            "sequential/g723_enc",
            "sequential/huff_enc",
            "sequential/statemate",
            "sequential/susan",
            "sequential/gsm_enc",
            "sequential/ndes",
            "sequential/audiobeam",
            "sequential/rijndael_enc",
            "sequential/cjpeg_transupp",
            "sequential/ammunition",
            "sequential/mpeg2",
            "sequential/anagram",
            "sequential/cjpeg_wrbmp",
            "sequential/petrinet",
        ]

    # @property
    # def skip_backend(self):
    #     return True

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("taclebench/", "")
            if name in self.supported_names:
                hint = TaclebenchProgram(
                    name,
                    alt=f"taclebench/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["TACLEBENCH_DIR"] = Path(self.config["taclebench.src_dir"])
        return ret

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "taclebench"
        return ret


class PolybenchFrontend(SimpleFrontend):
    REQUIRED = {"polybench.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "polybench",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        # TODO: automatic lookup
        return [
            "linear-algebra/solvers/gramschmidt",
            "linear-algebra/solvers/ludcmp",
            "linear-algebra/solvers/trisolv",
            "linear-algebra/solvers/durbin",
            "linear-algebra/solvers/lu",
            "linear-algebra/solvers/cholesky",
            "linear-algebra/kernels/atax",
            "linear-algebra/kernels/3mm",
            "linear-algebra/kernels/mvt",
            "linear-algebra/kernels/2mm",
            "linear-algebra/kernels/bicg",
            "linear-algebra/kernels/doitgen",
            "linear-algebra/blas/trmm",
            "linear-algebra/blas/gemver",
            "linear-algebra/blas/syrk",
            "linear-algebra/blas/gesummv",
            "linear-algebra/blas/syr2k",
            "linear-algebra/blas/symm",
            "linear-algebra/blas/gemm",
            "stencils/fdtd-2d",
            "stencils/seidel-2d",
            "stencils/adi",
            "stencils/jacobi-1d",
            "stencils/jacobi-2d",
            "stencils/heat-3d",
            "datamining/covariance",
            "datamining/correlation",
            "medley/deriche",
            "medley/nussinov",
            "medley/floyd-warshall",
        ]

    # @property
    # def skip_backend(self):
    #     return True

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("polybench/", "")
            if name in self.supported_names:
                hint = PolybenchProgram(
                    name,
                    alt=f"polybench/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["POLYBENCH_DIR"] = Path(self.config["polybench.src_dir"])
        return ret

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "polybench"
        return ret


class CoremarkFrontend(SimpleFrontend):
    REQUIRED = set()

    def __init__(self, features=None, config=None):
        super().__init__(
            "coremark",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        return [
            "coremark",
        ]

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("coremark/", "")
            if name in self.supported_names:
                hint = CoremarkProgram(
                    name,
                    alt=f"coremark/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "coremark"
        return ret


class DhrystoneFrontend(SimpleFrontend):
    REQUIRED = set()

    def __init__(self, features=None, config=None):
        super().__init__(
            "dhrystone",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        return [
            "dhrystone",
        ]

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("dhrystone/", "")
            if name in self.supported_names:
                hint = DhrystoneProgram(
                    name,
                    alt=f"dhrystone/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "dhrystone"
        return ret


class MathisFrontend(SimpleFrontend):
    REQUIRED = set()

    def __init__(self, features=None, config=None):
        super().__init__(
            "mathis",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        return [
            "to_upper",
            "add8",
            "add16",
            "gather_add8",
            "gather_add16",
            "scatter_add8",
            "scatter_add16",
            "dot8",
            "dot16",
            "saxpy8",
            "saxpy16",
            "matmul8",
            "matmul16",
            "matmul8_a",
            "matmul16_a",
            "transposed_matmul8",
            "transposed_matmul16",
            "transposed_matmul8_a",
            "transposed_matmul16_a",
            "transposed_matmul8_b",
            "transposed_matmul16_b",
        ]

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("mathis/", "")
            if name in self.supported_names:
                hint = MathisProgram(
                    name,
                    alt=f"mathis/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "mathis"
        return ret


class MibenchFrontend(SimpleFrontend):
    REQUIRED = {"mibench.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "mibench",
            ModelFormats.NONE,
            features=features,
            config=config,
        )

    @property
    def supported_names(self):
        # TODO: automatic lookup
        return [
            "telecomm/FFT",
            "telecomm/CRC32",
            "automotive/susan",
            "automotive/basicmath",
            "automotive/bitcount",
            "automotive/qsort",
            "security/sha",
            "security/rijndael",
            "network/dijkstra",
            "office/stringsearch",
        ]

    # @property
    # def skip_backend(self):
    #     return True

    def lookup_models(self, names, config=None, context=None):
        ret = []
        for name in names:
            name = name.replace("mibench/", "")
            if name in self.supported_names:
                hint = MibenchProgram(
                    name,
                    alt=f"mibench/{name}",
                    config=config,
                )
                ret.append(hint)
        return ret

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = [Artifact("dummy_model", raw=bytes(), fmt=ArtifactFormat.RAW, flags=["model", "dummy"])]

        return {"default": artifacts}, {}

    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["MIBENCH_DIR"] = Path(self.config["mibench.src_dir"])

        return ret

    def get_platform_config(self, platform):
        ret = {}
        if platform == "mlif":
            ret["template"] = "mibench"

        return ret


class LayerGenFrontend(Frontend):
    FEATURES = Frontend.FEATURES

    DEFAULTS = {
        **Frontend.DEFAULTS,
        "fmt": "tflite",  # TODO: relay
    }

    REQUIRED = Frontend.REQUIRED | {"layergen.exe"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "layergen",
            input_formats=[ModelFormats.TEXT],
            output_formats=[ModelFormats.TFLITE, ModelFormats.RELAY],
            features=features,
            config=config,
        )

    @property
    def fmt(self):
        value = self.config["fmt"]
        value = value.upper()
        assert value in ["TFLITE", "RELAY"]
        return value

    @property
    def layergen_exe(self):
        return Path(self.config["layergen.exe"])

    def produce_artifacts(self, model):
        pass

    # def produce_artifacts(self, model):
    #     artifacts = {}
    #     name = model.name
    #     path = model.paths[0]
    #     ext = ModelFormats[self.fmt].extension
    #     print("ext", ext)
    #     with open(path, "r") as handle:
    #         content = handle.read()
    #     lines = content.strip().split("\n")
    #     print("lines", lines, list(filter(None, lines)))
    #     assert len(lines) > 0, "Empty file not allowed."

    #     def helper(args):
    #         args = args.split(" ")
    #         with tempfile.TemporaryDirectory() as tmpdirname:
    #             out = Path(tmpdirname) / f"out.{ext}"
    #             utils.python(self.layergen_exe, self.fmt.lower(), out, *args, cwd=tmpdirname)
    #             # TODO: log output
    #             with open(out, "rb") as handle:
    #                 raw = handle.read()
    #             return raw

    #     if len(lines) > 1:
    #         for i, args in enumerate(lines):
    #             name = f"model{i}"
    #             raw = helper(args)
    #             artifact = Artifact(f"{name}.{ext}", raw=raw, fmt=ArtifactFormat.RAW, flags=["model"])
    #         artifacts[name] = [artifact]
    #     else:
    #         artifacts["default"] = []
    #     return artifacts

    def generate(self, model) -> Tuple[dict, dict]:
        artifacts = {}
        name = model.name
        path = model.paths[0]
        ext = ModelFormats[self.fmt].extension
        with open(path, "r") as handle:
            content = handle.read()
        lines = content.strip().split("\n")
        assert len(lines) > 0, "Empty file not allowed."

        def helper(args):
            args = args.split(" ")
            with tempfile.TemporaryDirectory() as tmpdirname:
                out = Path(tmpdirname) / f"out.{ext}"
                utils.python(self.layergen_exe, self.fmt.lower(), out, *args, cwd=tmpdirname)
                # TODO: log output
                with open(out, "rb") as handle:
                    raw = handle.read()
                return raw

        if len(lines) > 1:
            for i, args in enumerate(lines):
                name = f"model{i}"
                raw = helper(args)
                artifact = Artifact(f"{name}.{ext}", raw=raw, fmt=ArtifactFormat.RAW, flags=["model"])
                artifacts[name] = [artifact]
            # TODO: fix this
            artifacts["default"] = artifacts["model0"]  # Dummy model because default artifacts can not be empty
        else:
            name = "default"
            raw = helper(lines[0])
            artifact = Artifact(f"{name}.{ext}", raw=raw, fmt=ArtifactFormat.RAW, flags=["model"])
            artifacts[name] = [artifact]
        return artifacts, {}



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions of base classes for MLonMCU postprocesses."""
from mlonmcu.config import filter_config


class Postprocess:
    """Abstract postprocess."""

    FEATURES = set()

    DEFAULTS = {}

    REQUIRED = set()
    OPTIONAL = set()

    def __init__(self, name, config=None, features=None):
        self.name = name
        self.config = config if config else {}
        self.features = self.process_features(features)
        self.config = filter_config(self.config, self.name, self.DEFAULTS, self.OPTIONAL, self.REQUIRED)

    def process_features(self, features):
        """Utility which handles postprocess_features."""
        # Currently there is no support for postprocess features (FIXME)
        # if features is None:
        return []
        # features = get_matching_features(features, FeatureType.POSTPROCESS)
        # for feature in features:
        #     assert feature.name in self.FEATURES, f"Incompatible feature: {feature.name}"
        #     feature.add_target_config(self.name, self.config)
        # return features


class SessionPostprocess(Postprocess):
    """Session postprocess which is applied to multiple runs at the end of a session. (multi-row)"""

    def post_session(self, report):
        """Called at the end of a session."""


class RunPostprocess(Postprocess):
    """Run postprocess which is applied to a single run."""

    def post_run(self, report, artifacts):
        """Called at the end of a run."""



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import pkg_resources
import venv
import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_zephyr(context: MlonMcuContext, params=None):
    return context.environment.has_platform("zephyr")


# @Tasks.needs([])
@Tasks.provides(["zephyr.install_dir", "zephyr.sdk_dir", "zephyr.venv_dir"])
@Tasks.validate(_validate_zephyr)
@Tasks.register(category=TaskType.PLATFORM)
def install_zephyr(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install support for the Zephyr Platform."""
    zephyrName = utils.makeDirName("zephyr")
    zephyrInstallDir = context.environment.paths["deps"].path / "install" / zephyrName
    zephyrInstallDir.mkdir(exist_ok=True)
    zephyrSdkDir = zephyrInstallDir / "sdk"
    zephyrVenvDir = zephyrInstallDir / "venv"
    zephyrModulesDir = zephyrInstallDir / "modules"
    user_vars = context.environment.vars
    if "zephyr.install_dir" in user_vars:  # TODO: also check command line flags?
        assert "zephyr.sdk_dir" in user_vars
        assert "zephyr.venv_dir" in user_vars
        return False
    # boards = ["espressif"]
    # if "zephyr.boards" in user_vars:
    #     boards = user_vars["zephyr.boards"]
    # if not isinstance(boards, str):
    #     assert isinstance(boards, list)
    #     boards = ",".join(boards)
    if (
        not utils.is_populated(zephyrInstallDir)
        or not utils.is_populated(zephyrSdkDir)
        or not utils.is_populated(zephyrVenvDir)
        or not utils.is_populated(zephyrModulesDir)
        or rebuild
    ):
        zephyrVenvScript = zephyrVenvDir / "bin" / "activate"
        if not utils.is_populated(zephyrVenvDir):
            venv.create(zephyrVenvDir)
        utils.exec_getout(f". {zephyrVenvScript} && pip install west", shell=True, live=verbose)
        zephyrRepo = context.environment.repos["zephyr"]
        zephyrUrl = zephyrRepo.url
        if not utils.is_populated(zephyrInstallDir / "zephyr"):
            zephyrVersion = zephyrRepo.ref
            utils.exec_getout(
                f". {zephyrVenvScript} && west init -m {zephyrUrl} --mr {zephyrVersion} {zephyrInstallDir}",
                shell=True,
                live=verbose,
            )
        extra = zephyrInstallDir / "zephyr" / "scripts" / "requirements.txt"
        utils.exec_getout(f". {zephyrVenvScript} && pip install -r {extra}", shell=True, live=verbose)
        env = os.environ.copy()
        env["ZEPHYR_BASE"] = str(zephyrInstallDir / "zephyr")
        utils.exec_getout(f". {zephyrVenvScript} && west update", shell=True, live=verbose, env=env)
        if "zephyr.sdk_version" in user_vars:
            sdkVersion = user_vars["zephyr.sdk_version"]
        else:
            sdkVersion = "0.15.0-rc1"
        sdkDist = "linux-x86_64"
        sdkUrl = f"https://github.com/zephyrproject-rtos/sdk-ng/releases/download/v{sdkVersion}/"
        sdkArchive = f"zephyr-sdk-{sdkVersion}_{sdkDist}.tar.gz"
        if not utils.is_populated(zephyrSdkDir):
            utils.download_and_extract(sdkUrl, sdkArchive, zephyrSdkDir, progress=verbose)
        sdkScript = zephyrSdkDir / "setup.sh"
        # TODO: allow to limit installed toolchains
        utils.exec_getout(sdkScript, "-t", "all", "-h", live=verbose)
        # Apply patch to fix esp32c3 support
        patchFile = Path(
            pkg_resources.resource_filename(
                "mlonmcu", os.path.join("..", "resources", "patches", "zephyr", "fix_esp32c3_march.patch")
            )
        )
        if patchFile.is_file():
            xtensaDir = zephyrInstallDir / "modules" / "hal" / "xtensa"
            utils.patch(patchFile, cwd=xtensaDir)
    context.cache["zephyr.install_dir"] = zephyrInstallDir
    context.cache["zephyr.sdk_dir"] = zephyrSdkDir
    context.cache["zephyr.venv_dir"] = zephyrVenvDir



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import os
import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_cv32e40p(context: MlonMcuContext, params=None):
    return context.environment.has_target("cv32e40p")


@Tasks.provides(["cv32e40p.src_dir"])
@Tasks.validate(_validate_cv32e40p)
@Tasks.register(category=TaskType.TARGET)
def clone_cv32e40p(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the cv32e40p rtl."""
    name = utils.makeDirName("cv32e40p")
    srcDir = context.environment.paths["deps"].path / "src" / name
    user_vars = context.environment.vars
    # if "cv32e40p.verilator_executable" in user_vars:  # TODO: also check command line flags?
    if "cv32e40p.src_dir" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["cv32e40p"]
        utils.clone_wrapper(repo, srcDir, refresh=rebuild)
    context.cache["cv32e40p.src_dir"] = srcDir


@Tasks.needs(["cv32e40p.src_dir", "corevverif.src_dir", "verilator.install_dir"])
@Tasks.provides(["cv32e40p.verilator_executable", "cv32e40p.build_dir", "cv32e40p.install_dir"])
@Tasks.param("trace", [False, True])
@Tasks.param("waves", [False])
@Tasks.validate(_validate_cv32e40p)
@Tasks.register(category=TaskType.TARGET)
def build_cv32e40p(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build cv32e40p rtl."""
    if not params:
        params = {}
    trace = params.get("trace", False)
    waves = params.get("waves", False)
    user_vars = context.environment.vars
    if "cv32e40p.verilator_executable" in user_vars:  # TODO: also check command line flags?
        return False
    name = utils.makeDirName("cv32e40p")
    cv32e40pDir = Path(context.cache["cv32e40p.src_dir"])
    corevverifDir = Path(context.cache["corevverif.src_dir"])
    verilatorDir = Path(context.cache["verilator.install_dir"])
    cv32e40pSimDir = corevverifDir / "cv32e40p" / "sim" / "core"
    # buildDir = context.environment.paths["deps"].path / "build" / name
    buildDir = cv32e40pSimDir / "cobj_dir"
    installDir = context.environment.paths["deps"].path / "install" / name
    exe = installDir / "verilator_executable"
    if rebuild or not (utils.is_populated(buildDir) and exe.is_file()):
        # No need to build a vext and non-vext variant?
        utils.mkdirs(buildDir)
        args = [
            "CV_SW_TOOLCHAIN=_",
            "CV_SW_PREFIX=_",
            f"CV_CORE_PATH={cv32e40pDir}",
            f"CV_CORE_PKG={cv32e40pDir}",
            "VERBOSE=1",
        ]
        args.append(f"LOG_INSNS={int(trace)}")
        args.append(f"WAVES={int(waves)}")
        env = os.environ.copy()
        old_path = env["PATH"]
        env["PATH"] = f"{verilatorDir}/bin:{old_path}"
        utils.make("clean", *args, cwd=cv32e40pSimDir, threads=1, live=verbose, env=env)
        utils.make("testbench_verilator", *args, cwd=cv32e40pSimDir, threads=threads, live=verbose, env=env)
        utils.mkdirs(installDir)
        utils.move(cv32e40pSimDir / "simulation_results" / "hello-world" / "verilator_executable", exe)
    context.cache["cv32e40p.build_dir"] = buildDir
    context.cache["cv32e40p.install_dir"] = installDir
    context.cache["cv32e40p.verilator_executable"] = exe
    context.export_paths.add(installDir)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import numpy as np

from mlonmcu.models.utils import make_hex_array, fill_data_source, lookup_data_buffers, get_data_source


def test_models_utils_make_hex_array_bin(tmp_path_factory):
    data = [0, 1, 3, 5, 16, 33, 80]
    data = bytes(data)
    data_dir = tmp_path_factory.mktemp("data")
    data_file = data_dir / "f.bin"
    with open(data_file, "wb") as f:
        f.write(data)
    out = make_hex_array(data_file, mode="bin")
    assert out.strip() == "0x00, 0x01, 0x03, 0x05, 0x10, 0x21, 0x50,"
    out = make_hex_array(data_file, mode="auto")
    assert out.strip() == "0x00, 0x01, 0x03, 0x05, 0x10, 0x21, 0x50,"


@pytest.mark.parametrize("compressed", [False, True])
def test_models_utils_make_hex_array_npy(tmp_path_factory, compressed):
    data = np.array([0, 1, -3, 5, 16, -33, -80], dtype="int8")
    data_dir = tmp_path_factory.mktemp("data")
    ext = "npz" if compressed else "npy"
    data_file = data_dir / f"f.{ext}"
    if compressed:
        np.savez(data_file, data)
    else:
        np.save(data_file, data)
    out = make_hex_array(data_file, mode=ext)
    assert out.strip() == "0x00, 0x01, 0xfd, 0x05, 0x10, 0xdf, 0xb0,"
    out = make_hex_array(data_file, mode="auto")
    assert out.strip() == "0x00, 0x01, 0xfd, 0x05, 0x10, 0xdf, 0xb0,"


def test_models_utils_make_hex_array_invalid(tmp_path_factory):
    data = [0, 1, 3, 5, 16, 33, 80]
    data = bytes(data)
    data_dir = tmp_path_factory.mktemp("data")
    data_file = data_dir / "f"
    with pytest.raises(RuntimeError):
        make_hex_array(data_file, mode="foo")  # Invalid mode
    with pytest.raises(AssertionError):
        make_hex_array(data_file, mode="auto")  # Missing extension
    data_file = data_dir / "f.bin"
    with pytest.raises(FileNotFoundError):
        make_hex_array(data_file, mode="auto")  # File does not exist
    with open(data_file, "wb") as f:
        f.write(bytes([]))
    with pytest.raises(AssertionError):
        make_hex_array(data_file, mode="auto")  # Empty data


def test_models_utils_fill_data_source_empty():
    out = fill_data_source([], [])
    assert (
        out
        == """#include "ml_interface.h"
#include <stddef.h>
const int num_data_buffers_in = 0;
const int num_data_buffers_out = 0;
const unsigned char *const data_buffers_in[] = {};
const unsigned char *const data_buffers_out[] = {};
const size_t data_size_in[] = {};
const size_t data_size_out[] = {};
"""
    )


def test_models_utils_fill_data_source():
    out = fill_data_source([["0x01"], ["0x02"], ["0x03"]], [["0x04"], ["0x05"], ["0x06"]])
    assert (
        out
        == """#include "ml_interface.h"
#include <stddef.h>
const int num_data_buffers_in = 3;
const int num_data_buffers_out = 3;
const unsigned char data_buffer_in_0_0[] = {0x01};
const unsigned char data_buffer_in_1_0[] = {0x02};
const unsigned char data_buffer_in_2_0[] = {0x03};
const unsigned char data_buffer_out_0_0[] = {0x04};
const unsigned char data_buffer_out_1_0[] = {0x05};
const unsigned char data_buffer_out_2_0[] = {0x06};
const unsigned char *const data_buffers_in[] = {data_buffer_in_0_0, data_buffer_in_1_0, data_buffer_in_2_0, };
const unsigned char *const data_buffers_out[] = {data_buffer_out_0_0, data_buffer_out_1_0, data_buffer_out_2_0, };
const size_t data_size_in[] = {sizeof(data_buffer_in_0_0), sizeof(data_buffer_in_1_0), sizeof(data_buffer_in_2_0), };
const size_t data_size_out[] = {sizeof(data_buffer_out_0_0), sizeof(data_buffer_out_1_0), sizeof(data_buffer_out_2_0), };
"""  # noqa: E501
    )


@pytest.mark.parametrize("files", [False, True])
@pytest.mark.parametrize("fmt", ["bin", "npy", "npz", "foo"])
def test_models_utils_lookup_data_buffers_legacy(tmp_path_factory, files, fmt):
    ins_dir = tmp_path_factory.mktemp("ins")
    outs_dir = tmp_path_factory.mktemp("outs")

    def helper(fname):
        if fmt in ["npy", "npz"]:
            compressed = fmt == "npz"
            if compressed:
                np.savez(fname, np.ones(1, dtype="int8"))
            else:
                np.save(fname, np.ones(1, dtype="int8"))
        else:  # bin, foo
            with open(fname, "wb") as f:
                f.write(bytes([1]))

    in_names = ["0", "1", "2"]
    out_names = ["0", "1", "2"]

    in_files = [ins_dir / f"{name}.{fmt}" for name in in_names]
    out_files = [outs_dir / f"{name}.{fmt}" for name in out_names]

    list(map(helper, in_files))
    list(map(helper, out_files))

    if files:
        ins, outs = lookup_data_buffers(in_files, out_files)
    else:
        ins, outs = lookup_data_buffers([ins_dir], [outs_dir])
    assert len(ins) == len(outs)

    if fmt in ["bin", "npy", "npz"]:
        assert ins == [["0x01, "], ["0x01, "], ["0x01, "]]
        assert outs == [["0x01, "], ["0x01, "], ["0x01, "]]
    else:
        assert ins == []
        assert outs == []


def test_models_utils_lookup_data_buffers_legacy_multi(tmp_path_factory):
    ins_dir = tmp_path_factory.mktemp("ins")
    outs_dir = tmp_path_factory.mktemp("outs")

    def helper(fname):
        with open(fname, "wb") as f:
            f.write(bytes([1]))

    in_names = ["0_0", "0_1", "1_0", "1_1", "2_0", "2_1"]
    out_names = ["0_0", "0_1", "0_2", "1_0", "1_1", "1_2", "2_0", "2_1", "2_2"]

    in_files = [ins_dir / f"{name}.bin" for name in in_names]
    out_files = [outs_dir / f"{name}.bin" for name in out_names]

    list(map(helper, in_files))
    list(map(helper, out_files))

    ins, outs = lookup_data_buffers([ins_dir], [outs_dir])
    assert len(ins) == len(outs)

    assert ins == [["0x01, ", "0x01, "], ["0x01, ", "0x01, "], ["0x01, ", "0x01, "]]
    assert outs == [["0x01, ", "0x01, ", "0x01, "], ["0x01, ", "0x01, ", "0x01, "], ["0x01, ", "0x01, ", "0x01, "]]


# def test_models_utils_lookup_data_buffers(tmp_path_factory):
#     data_dir = tmp_path_factory.mktemp("data")
# def test_models_utils_lookup_data_buffers_multi(tmp_path_factory):
#     data_dir = tmp_path_factory.mktemp("data")


def test_models_utils_get_data_source():
    # empty
    out = get_data_source([], [])
    assert (
        out
        == """#include "ml_interface.h"
#include <stddef.h>
const int num_data_buffers_in = 0;
const int num_data_buffers_out = 0;
const unsigned char *const data_buffers_in[] = {};
const unsigned char *const data_buffers_out[] = {};
const size_t data_size_in[] = {};
const size_t data_size_out[] = {};
"""
    )

    # non empty
    # too complex



"""Test used RISC-V GCC Toolchains"""

# import pytest

# from mlonmcu.environment.config import PathConfig
# from mlonmcu.session.run import RunStage
# from mlonmcu.target.riscv.riscv import RISCVTarget
# from mlonmcu.target.riscv.riscv_vext_target import RVVTarget
# from mlonmcu.target.riscv.riscv_pext_target import RVPTarget
# from mlonmcu.target import register_target

# Using muRISCV-NN feature to get a relatively good coverage
# It would be nice if we would have specific test models which utilize all relevant kernels
# TODO: also test Zvl* extensions

# MODELS = ["tinymlperf"]
MODELS = ["resnet"]


# @pytest.mark.slow
# @pytest.mark.user_context
# @pytest.mark.needs
# @pytest.mark.parametrize("model_name", MODELS)
# @pytest.mark.parametrize("target", ["myriscv_default_llvm"])
# @pytest.mark.parametrize("xlen", [32, 64])
# # @pytest.mark.parametrize("fpu", ["none", "single", "double"])
# @pytest.mark.parametrize("fpu", ["none", "single", "double"])
# @pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
# @pytest.mark.parametrize("feature_names", [[], ["muriscvnn"]])
# def test_default(
#     model_name, target, xlen, fpu, embedded, compressed, atomic, multiply, feature_names, user_context, models_dir
# ):
#     has_multilib = _check_multilib(user_context, "riscv_gcc")
#     if xlen == 64 and not has_multilib:
#         pytest.skip(f"XLEN=64 needs multilib gcc")
#     if (not compressed or not atomic or not multiply) and not has_multilib:
#         pytest.skip(f"Non-multilib gcc only supports requires: c+m+a")
#     if embedded and not has_multilib:
#         pytest.skip(f"Embedded extension not supported by non-multilib gcc")
#     if fpu in ["none", "single"] and not has_multilib:
#         pytest.skip(f"Non-multilib gcc only supports double extension")
#
#     config = {
#         **_get_target_config(target, xlen, fpu, embedded, compressed, atomic, multiply),
#         **_get_platform_config(),
#     }
#
#     _, artifacts = _test_compile_platform(
#         "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config
#     )
#
#
# @pytest.mark.slow
# @pytest.mark.user_context
# @pytest.mark.parametrize("model_name", MODELS)
# @pytest.mark.parametrize("target", ["myriscv_vector_llvm"])
# # @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
# @pytest.mark.parametrize("vlen", [64, 128, 2048])
# @pytest.mark.parametrize("elen", [32])
# @pytest.mark.parametrize("spec", [1.0])
# # @pytest.mark.parametrize("fpu", ["none", "single", "double"])
# @pytest.mark.parametrize("fpu", ["none", "single"])
# @pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
# @pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
# def test_embedded_vector_32bit(
#     model_name,
#     target,
#     vlen,
#     elen,
#     spec,
#     fpu,
#     embedded,
#     compressed,
#     atomic,
#     multiply,
#     feature_names,
#     user_context,
#     models_dir,
# ):
#     has_multilib = _check_multilib(user_context, "riscv_gcc_vext")
#     if (not compressed or not atomic or not multiply) and not has_multilib:
#         pytest.skip(f"Non-multilib gcc only supports requires: c+m+a")
#     if embedded and not has_multilib:
#         pytest.skip(f"Embedded extension not supported by non-multilib gcc")
#     if fpu in ["none", "single"] and not has_multilib:
#         pytest.skip(f"Non-multilib gcc only supports double extension")
#
#     config = {
#         **_get_target_config(target, 32, fpu, embedded, compressed, atomic, multiply),
#         **_get_feature_vext_config(vlen, elen, spec, True),
#         **_get_platform_config(),
#     }
#
#     _, artifacts = _test_compile_platform(
#         "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config
#     )
#     _, artifacts = _test_compile_platform(
#         "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config
#     )
#
#
# @pytest.mark.slow
# @pytest.mark.user_context
# @pytest.mark.parametrize("model_name", MODELS)
# @pytest.mark.parametrize("target", ["myriscv_vector_llvm"])
# # @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
# @pytest.mark.parametrize("vlen", [64, 128, 2048])
# @pytest.mark.parametrize("elen", [32, 64])
# @pytest.mark.parametrize("spec", [1.0])
# @pytest.mark.parametrize("fpu", ["none", "single", "double"])
# @pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
# @pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
# def test_embedded_vector_64bit(
#     model_name,
#     target,
#     vlen,
#     elen,
#     spec,
#     fpu,
#     embedded,
#     compressed,
#     atomic,
#     multiply,
#     feature_names,
#     user_context,
#     models_dir,
# ):
#     has_multilib = _check_multilib(user_context, "riscv_gcc_vext")
#     if (not compressed or not atomic or not multiply) and not has_multilib:
#         pytest.skip(f"Non-multilib gcc only supports requires: c+m+a")
#     if embedded and not has_multilib:
#         pytest.skip(f"Embedded extension not supported by non-multilib gcc")
#
#     if elen == 32:
#         pytest.skip(f"ELEN 32 not supported due to compiler bug")
#     if fpu == "single":
#         pytest.skip(f"Single precision float not supported due to compiler bug")
#
#     config = {
#         **_get_target_config(target, 64, fpu, embedded, compressed, atomic, multiply),
#         **_get_feature_vext_config(vlen, elen, spec, True),
#         **_get_platform_config(),
#     }
#     _, artifacts = _test_compile_platform(
#         "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config
#     )
#
#
# # @pytest.mark.slow
# # @pytest.mark.user_context
# # @pytest.mark.parametrize("model_name", MODELS)
# # @pytest.mark.parametrize("target", ["myriscv_vector_llvm"])
# # # @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
# # @pytest.mark.parametrize("vlen", [64, 128, 2048])
# # @pytest.mark.parametrize("spec", [1.0])
# # # @pytest.mark.parametrize("fpu", ["none", "single", "double"])
# # @pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
# # @pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
# # def test_vector_32bit(model_name, target, vlen, spec, embedded, compressed,
#       atomic, multiply, feature_names, user_context, models_dir):
# #     config = {
# #         **get_target_config(target, 64, fpu, embedded, compressed, atomic, multiply),
# #         **get_feature_vext_config(vlen, 64, spec, True),
# #         **get_platform_config(),
# #     }
# #     _, artifacts = _test_compile_platform(
# #         "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config
# #     )
#
#
# @pytest.mark.slow
# @pytest.mark.user_context
# @pytest.mark.parametrize("model_name", MODELS)
# @pytest.mark.parametrize("target", ["myriscv_vector_llvm"])
# # @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
# @pytest.mark.parametrize("vlen", [64, 128, 2048])
# @pytest.mark.parametrize("spec", [1.0])
# # @pytest.mark.parametrize("fpu", ["none", "single", "double"])
# @pytest.mark.parametrize("fpu", ["double"])
# @pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
# @pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
# def test_vector_64bit(
#     model_name, target, vlen, spec, fpu, embedded, compressed, atomic, multiply,
#     feature_names, user_context, models_dir
# ):
#     has_multilib = _check_multilib(user_context, "riscv_gcc_vext")
#     if (not compressed or not atomic or not multiply) and not has_multilib:
#         pytest.skip(f"Non-multilib gcc only supports requires: c+m+a")
#     if embedded and not has_multilib:
#         pytest.skip(f"Embedded extension not supported by non-multilib gcc")
#
#     config = {
#         **_get_target_config(target, 64, fpu, embedded, compressed, atomic, multiply),
#         **_get_feature_vext_config(vlen, 64, spec, False),
#         **_get_platform_config(),
#     }
#     _, artifacts = _test_compile_platform(
#         "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config
#     )



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""TVM Target Platform"""
from mlonmcu.config import str2bool
from .tvm_rpc_platform import TvmRpcPlatform
from ..platform import TargetPlatform
from mlonmcu.target import get_targets
from mlonmcu.target.target import Target
from .tvm_target import create_tvm_platform_target
from mlonmcu.flow.tvm.backend.tvmc_utils import (
    get_bench_tvmc_args,
    get_data_tvmc_args,
    get_rpc_tvmc_args,
)


class TvmTargetPlatform(TargetPlatform, TvmRpcPlatform):
    """TVM target platform class."""

    FEATURES = (
        TargetPlatform.FEATURES
        | TvmRpcPlatform.FEATURES
        | {
            "benchmark",
            "tvm_profile",
        }
    )

    DEFAULTS = {
        **TargetPlatform.DEFAULTS,
        **TvmRpcPlatform.DEFAULTS,
        "fill_mode": None,  # random, zeros, ones, none
        "ins_file": None,
        "outs_file": None,
        "print_top": False,
        "profile": False,
        "repeat": 1,
        "number": 1,
        "aggregate": "none",  # Allowed: avg, max, min, none, all
        "total_time": False,
    }

    REQUIRED = TargetPlatform.REQUIRED | TvmRpcPlatform.REQUIRED

    @property
    def fill_mode(self):
        return self.config["fill_mode"]

    @property
    def ins_file(self):
        return self.config["ins_file"]

    @property
    def outs_file(self):
        return self.config["outs_file"]

    @property
    def print_top(self):
        return self.config["print_top"]

    @property
    def profile(self):
        value = self.config["profile"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def repeat(self):
        return self.config["repeat"]

    @property
    def number(self):
        return self.config["number"]

    @property
    def aggregate(self):
        value = self.config["aggregate"]
        assert value in ["avg", "all", "max", "min", "none"]
        return value

    @property
    def total_time(self):
        value = self.config["total_time"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def flash(self, elf, target, timeout=120):
        raise NotImplementedError

    def monitor(self, target, timeout=60):
        raise NotImplementedError

    def get_supported_targets(self):
        # TODO: get this via tvmc run --help
        target_names = ["cpu", "cuda", "cl", "metal", "vulkan", "rocm", "micro"]

        skip_names = ["micro"]  # Use microtvm platform instead

        return [f"tvm_{name}" for name in target_names if name not in skip_names]

    def create_target(self, name):
        assert name in self.get_supported_targets(), f"{name} is not a valid TVM device"
        targets = get_targets()
        if name in targets:
            base = targets[name]
        else:
            base = Target
        return create_tvm_platform_target(name, self, base=base)

    def get_tvmc_run_args(self):
        return [
            *get_data_tvmc_args(
                mode=self.fill_mode, ins_file=self.ins_file, outs_file=self.outs_file, print_top=self.print_top
            ),
            *get_bench_tvmc_args(
                print_time=True, profile=self.profile, end_to_end=False, repeat=self.repeat, number=self.number
            ),
            *get_rpc_tvmc_args(self.use_rpc, self.rpc_key, self.rpc_hostname, self.rpc_port),
        ]

    def invoke_tvmc_run(self, *args, target=None):
        assert target is not None, "Target required for tvmc run"
        combined_args = []
        combined_args.extend(["--device", target.device])
        return self.invoke_tvmc("run", *args)

    def run(self, elf, target, timeout=120):
        # TODO: implement timeout
        # Here, elf is actually a directory
        # TODO: replace workaround with possibility to pass TAR directly
        tar_path = str(elf)
        args = [tar_path] + self.get_tvmc_run_args()
        output = self.invoke_tvmc_run(*args, target=target)

        return output



import os
import logging

from mlonmcu.environment.environment import Environment
from mlonmcu.environment.config import DefaultsConfig, PathConfig, RepoConfig
from mlonmcu.environment.writer import create_environment_dict


class MyEnvironment(Environment):
    def __init__(self):
        super().__init__()
        self._home = "/foo/bar"
        self.defaults = DefaultsConfig(
            log_level=logging.DEBUG,
            log_to_file=False,
            log_rotate=False,
            default_framework=None,
            default_backends={},
            default_target=None,
            cleanup_auto=False,
            cleanup_keep=100,
        )
        self.paths = {"foo": PathConfig("bar"), "foobar": [PathConfig("baz"), PathConfig("baz2")]}
        self.repos = {"repo1": RepoConfig("repo1url"), "repo2": RepoConfig("repo2url", ref="repo2ref")}
        self.frameworks = []
        self.frontends = []
        self.toolchains = {}
        self.platforms = []
        self.targets = []
        self.vars = {"key": "val"}
        self.flags = {"my.var": ["foo", "bar"]}


def test_create_environment_dict():
    env = MyEnvironment()
    assert create_environment_dict(env) == {
        "home": "/foo/bar",
        "logging": {"level": "DEBUG", "to_file": False, "rotate": False},
        "cleanup": {"auto": False, "keep": 100},
        "paths": {
            "foo": os.path.join(os.getcwd(), "bar"),
            "foobar": [
                os.path.join(os.getcwd(), "baz"),
                os.path.join(os.getcwd(), "baz2"),
            ],
        },
        "repos": {
            "repo1": {"url": "repo1url", "ref": None, "options": {}},
            "repo2": {"url": "repo2url", "ref": "repo2ref", "options": {}},
        },
        "frameworks": {"default": None},
        "frontends": {},
        "toolchains": {},
        "platforms": {},
        "targets": {"default": None},
        "vars": {"key": "val"},
        "flags": {"my.var": ["foo", "bar"]},
    }



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class BackendModelOptions:
    def __init__(self, backend, supported=True, options={}):
        self.backend = backend
        self.supported = supported
        self.options = options


class TFLMIModelOptions(BackendModelOptions):
    def __init__(
        self,
        backend,
        supported=True,
        arena_size=None,
        builtin_ops=None,
        custom_ops=None,
    ):
        super().__init__(backend, supported=supported)
        self.arena_size = arena_size
        self.builtin_ops = builtin_ops
        self.custom_ops = custom_ops


class TVMRTModelOptions(BackendModelOptions):
    def __init__(self, backend, supported=True, arena_size=None):
        super().__init__(backend, supported=supported)
        self.arena_size = arena_size


def parse_model_options_for_backend(backend, options):
    backend_types = {
        "tflmi": TFLMIModelOptions,
        "tvmrt": TVMRTModelOptions,
    }
    if backend in backend_types:
        backend_type = backend_types[backend]
    else:
        backend_type = BackendModelOptions

    backend_options = backend_type(backend)

    for key, value in options.items():
        setattr(backend_options, key, value)

    return backend_options



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import io
import csv
import ast


class Metrics:
    def __init__(self):
        self.data = {}
        self.optional_keys = []
        self.order = []

    @staticmethod
    def from_csv(text):
        # TODO: how to find out data types? -> pandas?
        lines = text.splitlines()
        assert len(lines) == 2, "Metrics should only have two lines"
        # headers, data = lines[0].split(","), lines[1]
        empty = False
        for line in lines:
            if len(line.strip()) == 0:
                empty = True
        ret = Metrics()
        if not empty:
            reader = csv.DictReader(lines)
            data = list(reader)[0]
            data_ = {}
            for key, value in data.items():
                if len(key) > 2 and key[0] == key[-1] == "_":
                    new_key = key[1:-1]
                    data_[new_key] = value
                    ret.optional_keys.append(new_key)
                else:
                    data_[key] = value
            ret.data = data_
            ret.order = list(data_.keys())
        return ret

    def add(self, name, value, optional=False, overwrite=False, prepend=False):
        if not overwrite:
            assert name not in self.data, "Column with the same name already exists in metrics"
        self.data[name] = value
        if optional:
            self.optional_keys.append(name)
        if prepend:
            self.order.insert(0, name)
        else:
            self.order.append(name)

    def get(self, name):
        value = self.data[name]
        return (ast.literal_eval(value) if len(value) > 0 else None) if isinstance(value, str) else value

    def has(self, name):
        return name in self.data

    def get_data(self, include_optional=False, identify_optional=False):
        return {
            f"_{key}_" if identify_optional and key in self.optional_keys else key: self.get(key)
            for key in self.order
            if key not in self.optional_keys or include_optional
        }

    def to_csv(self, include_optional=False):
        data = self.get_data(include_optional=include_optional, identify_optional=True)
        output = io.StringIO()
        writer = csv.DictWriter(output, fieldnames=data.keys())
        writer.writeheader()
        writer.writerow(data)
        return output.getvalue()



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MicroTVM Platform"""

from mlonmcu.logging import get_logger

from .microtvm_base_platform import MicroTvmBasePlatform
from .microtvm_target_platform import MicroTvmTargetPlatform
from .microtvm_build_platform import MicroTvmBuildPlatform
from .microtvm_tune_platform import MicroTvmTunePlatform
from .microtvm_compile_platform import MicroTvmCompilePlatform

from .microtvm_target import create_microtvm_platform_target, get_microtvm_platform_targets

logger = get_logger()
# TODO: This file is very similar to the TVM platform -> Reuse as much as possible


class MicroTvmPlatform(
    MicroTvmBasePlatform,
    MicroTvmCompilePlatform,
    # MicroTvmTargetPlatform,  # implicitly via tune platform
    MicroTvmBuildPlatform,
    MicroTvmTunePlatform,
):
    """MicroTVM Platform class."""

    FEATURES = (
        MicroTvmBasePlatform.FEATURES
        | MicroTvmCompilePlatform.FEATURES
        | MicroTvmTargetPlatform.FEATURES
        | MicroTvmBuildPlatform.FEATURES
        | MicroTvmTunePlatform.FEATURES
    )  # TODO: validate?

    DEFAULTS = {
        **MicroTvmBasePlatform.DEFAULTS,
        **MicroTvmCompilePlatform.DEFAULTS,
        **MicroTvmTargetPlatform.DEFAULTS,
        **MicroTvmBuildPlatform.DEFAULTS,
        **MicroTvmTunePlatform.DEFAULTS,
    }

    REQUIRED = (
        MicroTvmBasePlatform.REQUIRED
        | MicroTvmCompilePlatform.REQUIRED
        | MicroTvmTargetPlatform.REQUIRED
        | MicroTvmBuildPlatform.REQUIRED
        | MicroTvmTunePlatform.REQUIRED
    )  # TODO: validate?

    def __init__(self, features=None, config=None):
        super(MicroTvmPlatform, self).__init__(
            "microtvm",
            features=features,
            config=config,
        )

    # The following methods are defined here as they would need to be implemented for compile and target platforms
    def get_supported_targets(self):
        return get_microtvm_platform_targets()

    def create_target(self, name):
        supported = self.get_supported_targets()
        assert name in supported, f"{name} is not a valid MicroTVM device"
        base = supported[name]
        return create_microtvm_platform_target(name, self, base=base)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import multiprocessing
import logging
import argparse

from mlonmcu.platform import get_platforms
from mlonmcu.session.postprocess import SUPPORTED_POSTPROCESSES
from mlonmcu.feature.features import get_available_feature_names
from mlonmcu.logging import get_logger, set_log_level
from .helper.parse import extract_config

logger = get_logger()

NUM_GEN_ARGS = 9


def handle_logging_flags(args):
    if hasattr(args, "verbose") and hasattr(args, "quiet"):
        if args.verbose and args.quiet:
            raise RuntimeError("--verbose and --quiet can not be used at the same time")
        elif args.verbose:
            set_log_level(logging.DEBUG)
        elif args.quiet:
            set_log_level(logging.WARNING)
        else:
            set_log_level(logging.INFO)


def add_flow_options(parser):
    flow_parser = parser.add_argument_group("flow options")
    flow_parser.add_argument(  # TODO: move to compile.py?
        "-t",
        "--target",
        type=str,
        metavar="TARGET",
        # choices=SUPPORTED_TARGETS.keys(),
        action="append",
        # default=None,
        # nargs=1,
        help="The target device/architecture (choices: See --list-targets)",
    )
    flow_parser.add_argument(  # TODO: move to compile.py?
        "--list-targets",
        action="store_true",
        help="List the supported targets in the environment",
    )
    flow_parser.add_argument(
        # "-p",
        "--platform",
        type=str,
        metavar="PLATFORM",
        choices=get_platforms().keys(),
        default=None,
        action="append",
        nargs=1,
        help="Explicitly choose the platforms to use (choices: %(choices)s)",
    )
    flow_parser.add_argument(
        # "-p",
        "--postprocess",
        type=str,
        metavar="POSTPROCESS",
        choices=SUPPORTED_POSTPROCESSES.keys(),
        action="append",
        help="Choose the postprocesses to apply (choices: %(choices)s)",
    )
    flow_parser.add_argument(
        "-f",
        "--feature",
        type=str,
        metavar="FEATURE",
        # nargs=1,
        action="append",
        choices=get_available_feature_names(),
        help="Enabled features for target/framework/backend (choices: %(choices)s)",
    )
    flow_parser.add_argument(
        "-c",
        "--config",
        metavar="KEY=VALUE",
        nargs="+",
        action="append",
        help=(
            "Set a number of key-value pairs "
            "(do not put spaces before or after the = sign). "
            "If a value contains spaces, you should define "
            "it with double quotes: "
            'foo="this is a sentence". Note that '
            "values are always treated as strings."
        ),
    )

    def add_gen_args(parser, number):
        for i in range(number):
            suffix = str(i + 1) if i > 0 else ""
            parser.add_argument(
                "--feature-gen" + suffix,
                dest="feature_gen" + suffix,
                type=str,
                metavar="FEATURES",
                nargs="+",
                action="append",
                help=(
                    f"Generator statement for features. (Also available: --feature-gen2 ... --feature-gen{number})"
                    if i == 0
                    else argparse.SUPPRESS
                ),
            )
            flow_parser.add_argument(
                "--config-gen" + suffix,
                dest="config_gen" + suffix,
                metavar="KEY=VALUE",
                nargs="+",
                action="append",
                help=(
                    f"Generator statement for configs. (Also available: --config-gen2 ... --config-gen{number})"
                    if i == 0
                    else argparse.SUPPRESS
                ),
            )

    add_gen_args(flow_parser, NUM_GEN_ARGS)
    flow_parser.add_argument(
        "--parallel",
        metavar="THREADS",
        nargs="?",
        type=int,
        const=multiprocessing.cpu_count(),
        default=1,
        help="Use multiple threads to process runs in parallel (%(const)s if specified, else %(default)s)",
    )
    flow_parser.add_argument(
        "-p",
        "--progress",
        action="store_true",
        help="Display progress bar (default: %(default)s)",
    )
    flow_parser.add_argument(
        "--resume",
        action="store_true",
        help="Try to resume the latest session (default: %(default)s)",
    )
    flow_parser.add_argument(  # TODO: move to compile.py?
        "-l",
        "--label",
        type=str,
        metavar="LABEL",
        default="",
        help="Label for the session (default: %(default)s)",
    )


def add_common_options(parser):
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print detailed messages for easier debugging (default: %(default)s)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Reduce number of logging statements to a minimum (default: %(default)s)",
    )
    parser.add_argument(
        "--docker",
        action="store_true",
        help="Execute run.py inside docker container (default: %(default)s)",
    )


def add_context_options(parser, with_home=True):
    common = parser.add_argument_group("context options")
    if with_home:
        home = os.getenv("MLONMCU_HOME", None)
        common.add_argument(
            "-H",
            "--home",
            "--hint",
            type=str,
            default=home,
            help="The path to the mlonmcu environment (overwriting $MLONMCU_HOME environment variable)",
        )


def add_model_options(parser):
    parser.add_argument(
        "models",
        metavar="model",
        type=str,
        nargs="+",
        default=None,
        help="Model to process",
    )


def kickoff_runs(args, until, context):
    assert len(context.sessions) > 0
    session = context.sessions[-1]
    # session.label = args.label
    config = extract_config(args)
    # TODO: move into context/session
    per_stage = True
    print_report = True
    if "runs_per_stage" in config:
        per_stage = bool(config["runs_per_stage"])
    elif "runs_per_stage" in context.environment.vars:
        per_stage = bool(context.environment.vars["runs_per_stage"])
    if "print_report" in config:
        print_report = bool(config["print_report"])
    elif "print_report" in context.environment.vars:
        print_report = bool(context.environment.vars["print_report"])
    with session:
        success = session.process_runs(
            until=until,
            per_stage=per_stage,
            print_report=print_report,
            num_workers=args.parallel,
            progress=args.progress,
            context=context,
            export=True,
        )
    if not success:
        logger.error("At least one error occured!")
        sys.exit(1)
